/*
 * Component: sr
 *
 * Description:
 *  Header file containing the definition of EventList and Event objects
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef EVENT_LIST_H_INCLUDED
#define EVENT_LIST_H_INCLUDED

#include <dcmtk/ofstd/ofstd.h>
#include <dcmtk/dcmdata/dctk.h>

#include "IMLogger.h"

/**
 * Event and EventList is used for collecting events which should
 * be sent to configured 3rd party devices.
 */


typedef enum {
	EVENT_NEW = 0,
	EVENT_STATUS = 1,
	EVENT_UPDATE = 2,
	EVENT_MERGE  = 3, // Study merged to order(s)
	EVENT_DELETE = 4
} EVENT_CODE;


/**
 * Abstract base class for handling study related events.
 */
class Event {
protected:

	/** Event code */
	EVENT_CODE m_eventCode;
	
	/** Study Instance UID the event belongs to */
	char * m_study;

	/** Status of the study at the time of the event */
	int m_status;

public:

	/**
	 * Constructor for creating a new Event object.
	 * This constructor is to be used for creating HL7 order events for a specified Scheduled Procedure Step (from MPPS).
	 *
	 * @param eventCode      Event code
	 * @param study          Study Instance UID the event belongs to
	 * @param status         Status of the study at the time of the event
	 */
	Event(EVENT_CODE eventCode, const char * study, int status);

	/**
	 * Destructor for cleaning up Event object
	 */
	virtual ~Event();

	/**
	 * @return event code
	 */
	inline EVENT_CODE getEventCode() {
		return m_eventCode;
	}

	/**
	 * @return Study Instance UID of the study the event belongs to
	 */
	inline const char * getStudy() {
		return m_study;
	}

	/**
	 * @return the status of the study
	 */
	inline int getStatus() {
		return m_status;
	}

	/**
	 * Send HL7 message triggered by the event.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if message was successfully sent or there was no message to be sent
	 *         false if an error occured
	 */
	virtual bool sendHL7Message(const char * deviceId) = 0;

	/**
	 * Send messages to the Scheduler module to be updated.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if message was successfully sent or there was no message to be sent
	 *         false if an error occured
	 */
	virtual bool sendToScheduler(const char * deviceId) = 0;
};


/**
 * Handling events of orders/studies.
 */
class OrderEvent : public Event {
private:

	/** Scheduled Procedure Step ID - trigger event only for the specified scheduled step in case of group orders */
	char * m_scheduledStep;

public:
	/**
	 * Constructor for creating a new Event object.
	 * This constructor is to be used for creating HL7 order events for a study/order.
	 *
	 * @param eventCode      Event code
	 * @param study          Study Instance UID the event belongs to
	 * @param status         Status of the study at the time of the event
	 */
	OrderEvent(EVENT_CODE eventCode, const char * study, int status);

	/**
	 * Constructor for creating a new Event object.
	 * This constructor is to be used for creating HL7 order events for a specified Scheduled Procedure Step (from MPPS).
	 *
	 * @param eventCode      Event code
	 * @param study          Study Instance UID the event belongs to
	 * @param scheduledStep  Scheduled Procedure Step ID - trigger event only for the
	 *                       specified scheduled step in case of group orders
	 * @param status         Status of the study at the time of the event
	 */
	OrderEvent(EVENT_CODE eventCode, const char * study, const char * scheduledStep, int status);

	/**
	 * Destructor for cleaning up OrderEvent object
	 */
	virtual ~OrderEvent();

	/**
	 * @return Scheduled Procedure Step ID
	 */
	inline const char * getScheduledStep() {
		return m_scheduledStep;
	}
	
	/**
	 * Send HL7 message triggered by the order event.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if message was successfully sent or there was no message to be sent
	 *         false if an error occured
	 */
	virtual bool sendHL7Message(const char * deviceId);

	/**
	 * Send messages to the Scheduler module to be updated.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if message was successfully sent or there was no message to be sent
	 *         false if an error occured
	 */
	virtual bool sendToScheduler(const char * deviceId);
};


/**
 * Handling events of study reports
 */
class ReportEvent : public Event {
private:	
	/** Name of the structured report file holding the report */
	char * m_srFile;
	
public:
	/**
	 * Constructor for creating a new Event object.
	 * This constructor is to be used for creating HL7 order events for a specified Scheduled Procedure Step (from MPPS).
	 *
	 * @param eventCode      Event code
	 * @param study          Study Instance UID the event belongs to
	 * @param srFile         Name of the Structured Report files holding the report content
	 * @param status         Status of the study at the time of the event
	 */
	ReportEvent(EVENT_CODE eventCode, const char * study, const char * srFile, int status);

	/**
	 * Destructor for cleaning up ReportEvent object
	 */
	virtual ~ReportEvent();

	/**
	 * @return Name of the Structured Report files holding the report content
	 */
	inline const char * getSrFile() {
		return m_srFile;
	}
	
	/**
	 * Send HL7 message triggered by the report event.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if message was successfully sent or there was no message to be sent
	 *         false if an error occured
	 */
	virtual bool sendHL7Message(const char * deviceId);

	/**
	 * Send messages to the Scheduler module to be updated.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if message was successfully sent or there was no message to be sent
	 *         false if an error occured
	 */
	virtual bool sendToScheduler(const char * deviceId);
};


#define INVALID_PB_STATUS -10000

/**
 * List of events.
 */
class EventList {
 private:

	/** List of events */
	OFList < Event * > m_eventList;

	/** Logger object for debug logging */
	IMLogger m_logger;

 public:

	/**
	 * Constructor for creating a new EventList object.
	 */
	EventList();


	/**
	 * Destructor for destroying and cleaning up this EventList object.
	 */
	~EventList();

	/**
	 * Adds an event to the event list.
	 *
	 * @param event   Event to be added to the event list
	 */
	inline void addEvent(Event * event) {
		m_eventList.push_back(event);
	}

	/**
	 * @return   iterator pointing to the first element in the event list
	 */
	inline OFIterator < Event * > getBeginIterator() {
		return m_eventList.begin();
	}

	/**
	 * @return  iterator pointing to the end of the event list
	 */
	inline OFIterator < Event * > getEndIterator() {
		return m_eventList.end();
	}

	/**
	 * Clears the event list.
	 * It is used to remove any other ORDER event when EVENT_MERGE is being added
	 * (when order completed message is sent no other update event should be sent)
	 */
	void clear();


	/**
	 * Add status changed event(s) to the event list.
	 *
	 * @param study       Study Instance UID of the study the events belongs to
	 * @param oldStatus   Old main status of the study
	 * @param newStatus   New main status of the study
	 */
	void addEventsForStatusChange(const char * study, int oldStatus, int newStatus);

	
	/**
	 * Send messages triggered by the events in the list.
	 *
	 * @param deviceId   device the events are originated from
	 *
	 * @return true if triggered message(s) was successfully sent
	 *         false if an error occured
	 */
	bool sendMessages(const char * deviceId);
};

#endif								  // EVENT_LIST_H_INCLUDED
