/*
 * Component: sr
 *
 * Description:
 *  Header file containing the definition of IMSRDocument class.
 *  Also contains the definition of ImageListElem and ImageList classes.
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef IMSRDOC_H_INCLUDED
#define IMSRDOC_H_INCLUDED

#include <iostream>				  /* for class ostream */

#include <dcmtk/dcmsr/dsrdoc.h>

#include "IMDocument.h"
#include "CodeTable.h"
// #include "EventList.h"
// #include "LogList.h"
// #include "PBItem.h"
#include "OrderDataFiller.h"


/**
 * Creating SR object with Blue Line information to be added to DICOM CDs to let eRAD viewer load the study quickly
 */
class IMBlueLineSRDocument : public DSRDocument, public IMDocument {
	private:

		/** Code manager used for creating CodedEntryValue objects */
		CodeManager * m_codeManager;

	protected:
		/**
		 * Creates a Unique Identifier in <code>uid</code> and returns <code>uid</code> for this document.
		 * <code>uid</code> must be at least 65 bytes. Care is taken to make sure that the generated UID is 64 characters or less.
		 *
		 * UID will be generated using IM_BLUE_LINE_SR_UID_ROOT prefix.
		 *
		 * @param uid    pointer to buffer of 65 or more characters in which the UID is returned
		 * @return pointer to UID, identical to <code>uid</code> parameter
		 */
		virtual char * generateUID( char * uid );

	public:

		/**
		 * Constructor for creating a new empty IMBlueLineSRDocument object
		 *
		 * @param codeManager   code manager used for creating CodedEntryValue objects
		 */
		IMBlueLineSRDocument( CodeManager * codeManager );

		/**
		 * Constructs a new <code>IMBlueLineSRDocument</code> for an existing blue line SR document.
		 *
		 * @param filename      File name of the existing SR document
		 * @param codeManager   code manager used for creating CodedEntryValue objects
		 */
		IMBlueLineSRDocument( const char * filename, CodeManager * codeManager );

		/**
		 * Destructs this <code>IMBlueLineSRDocument</code>
		 */
		virtual ~IMBlueLineSRDocument();

		/**
		 * Gets identifier of document type
		 * @return  <code>IMDocSR</code>
		 */
		virtual E_IMDocType getIdent() const {
			return IMDocBlueLine;
		}

		/**
		 * Loads the Blue Line Structured Report document file
		 *
		 * @param timezone  convert date/times in the dataset to this timezone
		 *                  (NULL: don't convert; "": convert to server local timezone)
		 *
		 * @throw IMException if an error occurs
		 */
		virtual void load( const char * timezone = NULL );

		/**
		 * Saves the Blue Line Structured Report document into file.
		 *
		 * @param filename  Save the document on this filename. If missing or NULL
		 *                  save on the filename the document was loaded or created.
		 *
		 * @throw IMException if an error occurs
		 */
		virtual void save( const char * filename = NULL );

		/**
		 * Creates new Blue Line Structured Report document for the given study.
		 * Creates SOP Instance UID and file name for the new document.
		 * Also creates Document Tree structure
		 *
		 * @param studyIUID   Study Instance UID of the study the document will belong to
		 *
		 * @throws IMException if an error occurs
		 */
		virtual void createNew( const char * studyIUID );

		OFBool createNew( const char * studyIUID, OrderDataFiller & odf, Sint32 seriesNumber /*, Sint32 instanceNumber*/ );

		OFCondition addBlueLineData( const char * fileName);

		/**
		 * Completes and verifies the SR document.
		 * The content of the approver sequence will be added to Verifying Observer sequence
		 *
		 * @param approverSequence  Sequence containing Verification information
		 *                         (verifier observer name and verification date/time)
		 *
		 * @throw IMException if an error occurs
		 */
		void completeAndVerifyDocument( DcmSequenceOfItems * approverSequence );

		inline void setStudyInstanceUID( const char * uid ) {
			if ( m_studyIUID != NULL )
				delete m_studyIUID;
			m_studyIUID = new char[ strlen( uid ) + 1 ];

			strcpy( m_studyIUID, uid );
		}

		inline void setSOPInstanceUID( const char * uid ) {
			if ( m_sopIUID != NULL )
				delete m_sopIUID;
			m_sopIUID = new char[ strlen( uid ) + 1 ];

			strcpy( m_sopIUID, uid );
		}

		inline DcmUniqueIdentifier *getSOPInstanceUIDElement() {
			return &SOPInstanceUID;
		}

		inline DcmUniqueIdentifier *getStudyInstanceUIDElement() {
			return &StudyInstanceUID;
		}

		/**
		 * Adds a new verifier (approver) to the Verifying Observer sequence
		 *
		 * @param observerName   Verifying Observer Name (Approver)
		 * @param observerCode   Verifying Observer Code (Not used)
		 * @param organization   Organization of the Observer (??/Not used)
		 * @param datetime       Date Time of Verification (Approval Date/Time)
		 *
		 * @return <UL><LI>EC_Normal    if verifier has been successfully added</LI>
		 *             <LI>EC_IllegalCall, EC_MemoryExhausted otherwise</LI></UL>
		 */
		OFCondition verifyDocument( const OFString & observerName, const DSRCodedEntryValue & observerCode,
				const OFString & organization, const OFString & datetime );


		void modifyDemography( DcmItem * item );


	private:

		/**
		 * Retrieves a top level (right under root) node.
		 *
		 * @param conceptName  identifies the container to be retrieved.
		 * @param baseNode     root node of the subtree to be searched (NULL means DocumentRoot)
		 *
		 * @return  pointer to the Node
		 *
		 * @throw IM_ASSERT exception if the root node was not found
		 */
		size_t findNode( DSRCodedEntryValue & conceptName, size_t baseNode = 0 );

		/**
		 * Retrieves a top level (right under root) node.
		 *
		 * @param conceptName        identifies the container to be retrieved.
		 * @param baseNode           root node of the subtree to be searched (NULL means DocumentRoot)
		 * @param relationshipType   type of relationship to parent node
		 * @param valueType          value type of the node
		 *
		 * @return  pointer to the Node
		 *
		 * @throw IM_ASSERT exception if the root node was not found
		 */
		size_t findNode( DSRCodedEntryValue & conceptName, const E_RelationshipType relationshipType, const E_ValueType valueType,
				size_t baseNode = 0 );

		size_t findOrCreateNode( DSRCodedEntryValue & conceptName, const E_RelationshipType relationshipType,
				const E_ValueType valueType, size_t baseNode = 0, size_t prevNode = 0 );

		/**
		 * Finds and retrieves a Text Tree Node with its Container Tree Node from the content tree.
		 * This method is used for retrieving impression and observation (findings, addendum) texts.
		 *
		 * @param containerName  Identifies the container tree node
		 * @param textName       Identifies the text tree node
		 * @param containerNode  Output parameter for the container node
		 * @param textNode       Output parameter for the text node
		 */
		void getTextWithContainer( DSRCodedEntryValue & containerName, DSRCodedEntryValue & textName, size_t & containerNode,
				size_t & textNode );

		/**
		 * Finds and retrieves a Text Tree Node with its Container Tree Node from the content tree.
		 * The existence of the text and container tree nodes will be checked.
		 * This method is used for retrieving impression and observation (findings, addendum) texts.
		 *
		 * @param containerName  Identifies the container tree node
		 * @param textName       Identifies the text tree node
		 * @param containerNode  Output parameter for the container node
		 * @param textNode       Output parameter for the text node
		 *
		 * @throw IMException if either the container node or the text node is missing.
		 */
		void getAndCheckTextWithContainer( DSRCodedEntryValue & containerName, DSRCodedEntryValue & textName, size_t & containerNode,
				size_t & textNode );

		/**
		 * Sets string value of the current tree node in the content tree
		 *
		 * @param strValue   the new value of the tree node
		 *
		 * @return <UL><LI>OFTrue   if the value of the current node has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool setCurrentStringValue( const char * strValue );

		/**
		 * Sets coded entry value of the current tree node in the content tree
		 *
		 * @param codedEntryValue   the new value of the tree node
		 *
		 * @return <UL><LI>OFTrue   if the value of the current node has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool setCurrentCodeValue( const DSRCodedEntryValue & codedEntryValue );

		/**
		 * Sets string value of the current tree node in the content tree. The new value of the
		 * tree node will be read from a file
		 *
		 * @param filename   name of the file containing the new value for the current tree node
		 *
		 * @return <UL><LI>OFTrue   if the value of the current node has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyStringFromFile( const char * filename );

		void checkAndFixIDs( DcmItem & item, DcmTagKey idTag, const char * uidPrefix, OFString & origSop, OFString & sop,
				OFString & origId, OFString & id );
		void addInstanceCreationAttributes( DcmItem & item );

		/**
		 * Finds and retrieves the string value of an element from an item
		 *
		 * @param src    the item to search for the element
		 * @param key    identifies the element to find
		 * @param value  output parameter for the element value
		 *
		 * @return <UL><LI>OFTrue  if the element and its value has been successfully retrieved</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool getElementValue( DcmItem * src, DcmTagKey key, const char *& value );

		/**
		 * Checks if the given tag is a demography tag or not
		 *
		 * @param tag     the tag to check
		 *
		 * @return OFTrue  if and only if the tag is NOT a demography tag
		 */
		OFBool isNotDemographyTag( DcmTagKey & tag );

		/**
		 * Reads SR content tree (and SR related tags) from dataset
		 * Also reads results and interpretation sequences
		 *
		 * @param dataset  contains the SR content tree
		 * @param flags    flags applied while reading (see SRDocument::read())
		 *
		 * @return EC_Normal  if content has been successfully read or error code
		 * @see SRDocument::read()
		 *
		 * @throw IMException if an error occurs
		 */
		OFCondition read( DcmItem & dataset, const size_t flags );

		/**
		 * Writes SR content tree (and SR related tags) into dataset
		 * Also writes results and interpretation sequences
		 *
		 * @param dataset      write SR content tree into this dataset
		 * @param markedItems  (see SRDocument::write())
		 *
		 * @return EC_Normal  if content has been successfully written or error code
		 * @see SRDocument::write()
		 *
		 * @throw IMException if an error occurs
		 */
		OFCondition write( DcmItem & dataset, DcmStack * markedItems );

		/**
		 * Modifies element in the dataset.
		 *
		 * @param elem        identifies the element to change and contains new element value
		 * @param createOnly  if OFTrue then existing element will not be modified
		 *
		 * @return <UL><LI>OFTrue  value of element has been changed</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyElement( DcmElement * elem, OFBool createOnly = OFFalse );

		/**
		 * Modifies element in the first item of the sequence.
		 *
		 * @param sequence    modify element in the first item of this sequence
		 * @param elem        identifies the element to change and contains new element value
		 * @param createOnly  if OFTrue then existing element will not be modified
		 *
		 * @return <UL><LI>OFTrue  value of element has been changed</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyElementInSequence( DcmSequenceOfItems * sequence, DcmElement * elem, OFBool createOnly );

		/**
		 * Modifies the (string) value of an element by copying the value of an another element
		 *
		 * @param elem     the element to modify
		 * @param modElem  copy the value of this element into elem
		 *
		 * @return OFTrue  if and only if the value of elem has been changed
		 */
		OFBool modifyElemValue( DcmElement & elem, DcmElement * modElem );

		IMBlueLineSRDocument( const IMBlueLineSRDocument & );
		IMBlueLineSRDocument & operator=( const IMBlueLineSRDocument & );
};

#endif

