/*
 * Component: sr
 *
 * Description:
 *  Header file containing the definition of IMDocument object
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef IM_DOCUMENT_H_INCLUDED
#define IM_DOCUMENT_H_INCLUDED

#include <time.h>

#include <dcmtk/dcmdata/dctk.h>

#include <PBItem.h>

/**
 * Supported Document types
 */
typedef enum {
	/** PbR Document */
	IMDocPbR = 0,
	/** PbD Document */
	IMDocPbD = 1,
	/** SR Document */
	IMDocSR = 2,
	/** Image (SC) object */
	IMDocImg = 3,
	/** Encapsulated PDF document */
	IMDocPdf = 4,
	/** Blue Line SR Document */
	IMDocBlueLine = 5
} E_IMDocType;

/**
 * Base class for Document types created by EP software
 */
class IMDocument {
 private:

	/**
	 * Initialize IM Document
	 */
	void initDocument();

 protected:

	static DcmTagKey m_supportedStudyAttributes[];
	static const int m_numberOfSupportedStudyAttributes;

	/** File format holding the DICOM object representing the IM Document */
	DcmFileFormat * m_fileformat;

	/** DICOM dataset of the DICOM object */
	DcmDataset * m_dataset;

	/** Path to the DICOM object */
	char * m_filename;

	/** Short file name of the DICOM object */
	OFString m_shortName;

	/** Study Instance UID of the study the object belongs to */
	char * m_studyIUID;

	/** SOP Instance UID of the DICOM object */
	char * m_sopIUID;

	/** Indicates whether the DICOM object was changed and to be saved */
	OFBool m_isChanged;

	/** Indicates whether registration tasks should be started directly instead of via taskd */
	OFBool m_registerDirectly;

	/** Indicates whether registration tasks were started from obsolete tool */
	OFBool m_startedFromObsolete;

	/** Offset from UTC */
	OFString m_offsetFromUTC;

	/** true indicates that offset from UTC was already read from the dataset */
	OFBool m_wasOffsetFromUTCInitialized;

	/** true indicates this is a newly created document */
	OFBool m_isNew;

	/**
	 * Compares two string values.
	 * Helper method for calcLastModifiedDateTime()
	 *
	 * @param val1   string value 1 to compare
	 * @param val2   string value 2 to compare
	 *
	 * @return an integer  less  than,  equal to,  or  greater  than zero if val1
	 *         is found respectively to be less than, to match, or be greater than val2
	 */
	int compare( const char * val1, const char * val2 );

	/**
	 * Reads configured character set to be set in DICOM objects generated by EP
	 *
	 * @return configured character set or 'ISO_IR 100' as the built in default
	 */
	OFString getConfiguredCharset();

	/**
	 * Creates a Unique Identifier in <code>uid</code> and returns <code>uid</code> for this document.
	 * <code>uid</code> must be at least 65 bytes. Care is taken to make sure that the generated UID is 64 characters or less.
	 *
	 * Default implementation generates UID using IM_INSTANCE_UID_ROOT prefix.
	 *
	 * @param uid    pointer to buffer of 65 or more characters in which the UID is returned
	 * @return pointer to UID, identical to <code>uid</code> parameter
	 */
	virtual char * generateUID( char * uid );


	/**
	 * Sets PB Acquisition Date and Time tags in a PB item.
	 * PB Acquisition Date and Time tags shall be updated with current time in UTC
	 *
	 * @param item     Sets PB Acquisition Date/Time in this PB item
	 * @param replace  OFTrue indicates to overwrite existing value (default)
	 */
	void setPBAcquisitionDateTime( PBItem & item, OFBool replace = OFTrue );


 public:

	/**
	 * Constructor for creating a new empty IMDocument object.
	 */
	IMDocument();

	/**
	 * Constructor for creating a new IMDocument object for an existing document.
	 *
	 * @param filename   File name of the existing document
	 */
	IMDocument(const char * filename);

	/**
	 * Destructor for destroying an IMDocument object.
	 */
	virtual ~IMDocument();


	/**
	 * @return  identifier of document type
	 */
	virtual E_IMDocType getIdent() const = 0;

	/**
	 * Loads the Image Medical Document file
	 *
	 * @param timezone  convert date/times in the dataset to this timezone
	 *                  (NULL: don't convert; "": convert to server local timezone)
	 *
	 * @throw IMException if an error occurs
	 */
	virtual void load( const char * timezone = NULL );

	/**
	 * Saves the Image Medical Document into file.
	 *
	 * @param filename  Save the document on this filename. If missing or NULL
	 *                  save on the filename the document was loaded or created.
	 *
	 * @throw IMException if an error occurs
	 */
	virtual void save( const char * fileName = NULL );

	/**
	 * Saves the document in IMAGE_TMP and starts registering it to the database
	 *
	 * @return OFTrue   If saving the document and starting the registration to the database was successful
	 *          OFFalse  Otherwise
	 *
	 * @throw IMException if an error occurs
	 */
	OFBool saveAndRegisterDocument();

	/**
	 * Sets Image Medical specific tags and values in the Document
	 */
	virtual void setIMTags();

	/**
	 * Remove this Image Medical Document from file system.
	 *
	 * @return OFTrue   if the document was successfully removed
	 *          OFFalse  otherwise
	 */
	OFBool remove();

	/**
	 * Creates new Image Medical Document for the given study.
	 * Creates SOP Instance UID and file name for the new document.
	 *
	 * @param studyIUID   Study Instance UID of the study the document will belong to
	 *
	 * @throw IMException if an error occurs
	 */
	virtual void createNew( const char * studyIUID );

	/**
	 * Creates a backup copy of Image Medical Document in the given directory.
	 *
	 * @param dirName   Name of the backup directory
	 *
	 * @throw IMException if an error occurs
	 */
	void backup( const char * dirName );

	/**
	 * Creates a backup copy of Image Medical Document with the given filename.
	 *
	 * @param backupFileName   Name of the backup file to be created
	 *
	 * @throw IMException if an error occurs
	 */
	void backupFile( const char * backupFileName );

	/**
	 * Checks if the Image Medical Document was created by the recent software version
	 *
	 * @return OFTrue   If the Document was created by the recent software version
	 *          OFFasle  Otherwise
	 */
	OFBool isLatestVersion();

	/**
	 * Gets last modified date and time of the Image Medical Document.
	 *
	 * @returns   last modified date and time of the Document file
	 */
	time_t getLastModified();

	/**
	 * Retrieves last modification date and time of the Image Medical Document stored in
	 * PB Last Modified Date and PB Last Modified Time tags
	 *
	 * @param lmdate   Output parameter for last modified date
	 * @param lmtime   Output parameter for last modified time
	 */
	void getLastModifiedDateTime( const char *& lmdate, const char *& lmtime );

	/**
	 * Calculates newest last modified date/time of several documents.
	 * This method will be called for each Document when last modified date/time are to be calculated.
	 *
	 * @param mainDate   in/out parameter for last modified date
	 * @param mainTime   in/out parameter for last modified time
	 * @param curDate    last modified date of the current Document
	 * @param curTime    last modified time of the current Document
	 *
	 * @return true if mainDate or mainTime was changed (curDate/curTime was newer); false otherwise
	 */
	bool calcLastModifiedDateTime( const char *& mainDate, const char *& mainTime, const char * curDate, const char * curTime );

	/**
	 * Retrieves the SOP Instance UID of the Image Medical Document.
	 *
	 * @param documentUID   Output parameter for SOP Instance UID
	 */
	inline void getDocumentUID( const char *& documentUID ) {
		m_dataset->findAndGetString( DCM_SOPInstanceUID, documentUID );
	}

	/**
	 * @return  path to the IM Document object file
	 */
	inline const char * getFileName() const {
		return m_filename;
	}

	/**
	 * @return short name of the document file
	 */
	inline const OFString & getShortName() const {
		return m_shortName;
	}

	/**
	 * @return File format holding the DICOM object representing the IM Document
	 */
	inline DcmFileFormat * getFileFormat() {
		return m_fileformat;
	}

	/**
	 * @return Dataset of the IM Document
	 */
	inline DcmDataset * getDataset() {
		return m_dataset;
	}

	/**
	 * @return  OFTrue if and only if the document has been changed and need to be saved
	 */
	inline OFBool isChanged() const {
		return m_isChanged;
	}

	/**
	 * Sets the changed flag for this document.
	 *
	 * @param isChanged  New change value - will be 'or'-ed to the current changed flag
	 */
	inline void setChanged( OFBool isChanged ) {
		m_isChanged = m_isChanged || isChanged;
	}

	/**
	 * @return TimezoneOffsetFromUTC set in the dataset (or empty string if offset is not set)
	 */
	inline OFString & getOffsetFromUTC() {
		if ( !m_wasOffsetFromUTCInitialized ) {
			m_dataset->findAndGetOFString( DCM_TimezoneOffsetFromUTC, m_offsetFromUTC );
			m_wasOffsetFromUTCInitialized = OFTrue;
		}
		return  m_offsetFromUTC;
	}

	/**
	 * @return true if and only if the document was just created
	 */
	inline OFBool isNew() const {
		return m_isNew;
	}

	/**
	 * Copies supported study and patient demography tags from the provided dataset to the document dataset.
	 *
	 * @param dataset     to get the study and patient demography data from
	 * @param replaceOld  true indicates to replace existing data if any (default: OFTrue)
	 *
	 * @return true if and only if study and patient demography data was successfully added to the document dataset
	 */
	virtual bool copySupportedTags( DcmDataset * dataset, OFBool replaceOld = OFTrue );
};

#endif								  // IM_DOCUMENT_H_INCLUDED
