/*
 * Component: sr
 *
 * Description:
 *  Header file containing the definition of IMPbRDocument object
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef IM_PBR_DOCUMENT_H_INCLUDED
#define IM_PBR_DOCUMENT_H_INCLUDED

// #include <dcmtk/config/osconfig.h>
#include <dcmtk/dcmdata/dctk.h>

#include "PBCustomFields.h"

#include "IMDocument.h"
#include "IMPbDDocument.h"
#include "PBItem.h"
#include "dtconv.h"
#include "IMLogger.h"

#define T_UID_LEN 65
typedef char T_UID[ T_UID_LEN ];

/**
 * Helper class to hold SOP Class UID and SOP Instance UID pairs as SOP Instance Reference identifiers.
 */
class SOPInstanceReference {
	private:
		OFString m_sopClass;
		OFString m_sopInstance;

	public:

		/**
		 * Constructs a new SOP Instance Reference object
		 * @param sopClass     SOP Class UID of the object to be referenced
		 * @param sopInstance  SOP Instance UID of the object to be referenced
		 */
		SOPInstanceReference( const OFString & sopClass, const OFString & sopInstance ) {
			m_sopClass = sopClass;
			m_sopInstance = sopInstance;
		}

		/**
		 * Copy constructor
		 *
		 * @param orig original SOP Instance Reference object to be copied
		 */
		SOPInstanceReference( const SOPInstanceReference & orig ) {
			m_sopClass = orig.m_sopClass;
			m_sopInstance = orig.m_sopInstance;
		}

		/**
		 * @return SOP Class UID of the referenced object
		 */
		inline const OFString & getSOPClass() const {
			return m_sopClass;
		}

		/**
		 * @return SOP Instance UID of the referenced object
		 */
		inline const OFString & getSOPInstance() const {
			return m_sopInstance;
		}

		/**
		 * = operator
		 * @param orig  original SOP Instance Reference object
		 */
		SOPInstanceReference & operator= (const SOPInstanceReference & orig) {
			m_sopClass = orig.m_sopClass;
			m_sopInstance = orig.m_sopInstance;
			return *this;
		}
};

class DcmSeeker; // Defined in utils/dcmutils

class IMPbRDocument : public IMDocument {
	private:
		static IMLogger m_logger;

		IMPbDDocument * m_pbdDocument;
		OFBool m_force;

		OFBool m_updatePBAcquisitionDateTime;

		// Commonly used functions
		OFBool isValidFigure( DcmElement * figure );
		void createElementInItem( DcmItem * item, DcmTagKey key, const char * value );
		void createItemSOPIUID( char *fileSOPIUID, Uint16 privateElem, T_UID & sopiuid );
		DcmItem * createItem( char * fileSOPIUID, Uint16 privateElem );
		void addSequence( DcmDataset * dataset, Uint16 privateElem, DcmItem * item );
		DcmItem * createSeqAndItem( DcmDataset * dataset, char * fileSOPIUID, Uint16 privateElem );
		int statusIDtoPbRStatus( DcmElement * statusID );
		void createReferences( DcmItem * interpretationItem, DcmItem * dictationItem );

		// Functions used for canonizing old style PbRs

		//  size_t URLDecode(void *pString); // also exists in helper.h
		size_t DecodeBinary( void * pString, size_t nString, void * pBinary );
		OFBool convertOldDictation( DcmElement * dictation, DcmItem * interpretationItem );
		void oldStyle2v1();

		// Functions used for canonizing v0 PbRs
		OFBool isInterpretationElement( DcmTagKey & key, PBItem & datasetPBItem, DcmObject * obj );
		OFBool isResultsElement( DcmTagKey & key, PBItem & datasetPBItem, DcmObject * obj );
		void convertV0Dictation( DcmItem * interpretationItem );
		void v0tov1();

		// Functions used for canonizing v1 PbRs
		OFBool checkResultsItem( DcmItem * resultsItem );
		OFBool checkAndFixResultsSequence( DcmSequenceOfItems * resultsSeq );
		OFBool checkInterpretationItem( DcmItem * interpretationItem, int & statusID );
		OFBool checkAndFixInterpretationSequence( DcmSequenceOfItems * interpretationSeq, int & statusID );
		void canonizeVersion1PbR();

		OFBool processSeriesModificationSequence( DcmSequenceOfItems * seriesSequence, DcmSequenceOfItems * modSequence );
		OFBool processSeriesModificationItem( DcmItem * seriesItem, DcmItem * modItem );
		OFBool processInstanceModificationSequence( DcmSequenceOfItems * instanceSequence, DcmSequenceOfItems * modSequence );
		OFBool processInstanceModificationItem( DcmItem * instanceItem, DcmItem * modItem );

		DcmItem * getStudyModificationItem( OFBool create = OFFalse );
		DcmItem * getItem( DcmSequenceOfItems * sequence, const char * itemuid, OFBool create = OFFalse );

		static OFBool copyElem( DcmSeeker & seeker, const DcmTagKey & srcKey, DcmItem & item, const DcmTagKey & destKey,
				OFBool isMove );
		static void copyElemIntoItem( DcmElement & elem, DcmItem & item, const DcmTagKey & newKey );

		static DcmItem * passCodeItem( DcmItem * item );
		static OFBool processCodeSequence( DcmSequenceOfItems * pbrCodeSeq, DcmSequenceOfItems * seq );

		static OFBool isTheSameCodeSequence( DcmSequenceOfItems * seq1, DcmSequenceOfItems * seq2 );
		static OFBool isTheSameCodeItem( DcmItem * item1, DcmItem * item2 );
		void checkAndFillCPTandICDCodes();

		/**
		 * Checks whether the PbR belongs to a study (vs. order; ie. the status in the dataset is greater than 0)
		 * or would become a study after modifications applied.
		 *
		 * @param modItem modification item
		 *
		 * @return true if and only if the PbR after save would belong to a study
		 */
//		OFBool isStudyPbR( PBItem & pbrItem, PBItem & modItem ) {

//		bool isStudyPbR( const PBItem & modItem );
	public:

		IMPbRDocument();
		IMPbRDocument( const char * filename, IMPbDDocument * dictation );
		virtual ~IMPbRDocument();

		virtual E_IMDocType getIdent() const {
			return IMDocPbR;
		}
		void createNew( const char * studyIUID );

		virtual void save( const char * fileName = NULL );

		OFBool canonize();

		OFBool processStudyModificationItem( DcmItem * item );
		OFBool processCustomFields( const PBCustomFields & customFields );

		inline void forceConversion() {
			m_force = OFTrue;
		}

		/**
		 * Sets Image Medical specific tags and values in the Document
		 */
		virtual void setIMTags();

		/**
		 * Sets the value of the register directly flag (which defaults to OFTrue for PbR documents)
		 * @param newValue   new value for the register directly flag
		 */
		inline void setRegisterDirectly( OFBool newValue ) {
			m_registerDirectly = newValue;
		}

		/**
		 * Sets the value of the started from obsolete tool flag (which defaults to OFFalse)
		 * @param newValue	new value for the started from obsolete tool flag
		 */
		inline void setStartedFromObsolete( OFBool newValue ) {
			m_startedFromObsolete = newValue;
		}

		void setStatus( const char * strStatus );

		OFBool containsValidReport();

		void copyRequestData( DcmItem & requestItem, OFBool replaceOldValues = OFTrue, OFBool isMove = OFFalse );
		void copyPBProcedureData( DcmSequenceOfItems & procedureSeq, OFBool replaceOldValues = OFTrue );
		void copyPBPerformedProcedureData( DcmSequenceOfItems & destPerformedSeq, OFBool replaceOldValues = OFTrue );

		OFBool editPBProcedureData( DcmSequenceOfItems & procedureSeq );

		void applyModificationDataset( pbtz::OffsetAdjustedDataset & ofDataset, OFBool createOnly = OFFalse, DcmElement ** status = NULL );

		/**
		 * Sets the whole study obsolete.
		 * Adds (or modifies) PB_ObsoleteStatus in the PbR dataset. the value of the tag will be set to 'STUDY'
		 *
		 * @return  true if the entire study in the PbR object was successful
		 */
		OFBool setStudyObsolete();

		/**
		 * Invalidates objects acquired before given date/time in the study.
		 * Adds (or modifies) PB_InvalidateDate and PB_InvalidateTime tags in the PbR dataset
		 * Also sets the PB_ObsoleteStatus to 'PARTIAL' in the PbR dataset - if not yet set
		 * If seriesUID is provided, only the given series shall be invalidated
		 *
		 * @param date       value for the PB_InvalidateDate tag
		 * @param time       value for the PB_InvalidateTime tag
		 * @param seriesUID  UID of the series to invalidate (NULL means entire study; default is NULL)
		 *
		 * @return  true if invalidate date/time was successfully set or the study was already obsolete. Returns false otherwise.
		 */
		OFBool invalidate( const OFDate & date, const OFTime & time, const char * seriesUID = NULL );

		/**
		 * Set objects obsolete in the study.
		 * Adds (or modifies) PB_ObsoletedObjectSequence to the PbR dataset
		 * Also sets the PB_ObsoleteStatus to 'PARTIAL' in the PbR dataset - if not yet set
		 *
		 * @param files   list of references to the objects to be set obsolete
		 *
		 * @return  true if objects were successfully set obsolete or the study was already obsolete. Returns false otherwise.
		 */
		OFBool setObjectsObsolete( OFList<SOPInstanceReference> & files );

		static OFBool copyRequestData( DcmItem & src, DcmItem & dest, OFBool replaceOldValues = OFTrue, OFBool isMove = OFFalse );
		static DcmItem * findOrCreateItem( DcmSequenceOfItems & seq, DcmItem & item, const DcmTagKey & idKey, bool returnFirst =
				false );
		static DcmSequenceOfItems * findOrCreateSequence( DcmItem & item, DcmSequenceOfItems & seq );
		static void copyScheduledData( DcmSequenceOfItems & scheduledSeq, DcmItem & item, OFBool replaceOldValues = OFTrue );
		static void copyPerformedData( DcmSequenceOfItems & performedSeq, DcmItem & item, OFBool replaceOldValues = OFTrue );
		static void mergeReferencedObjectSequences( DcmSequenceOfItems & dest, DcmSequenceOfItems & src );
		static void copySeriesData( DcmSequenceOfItems & seriesSeq, DcmItem & item, OFBool replaceOldValues = OFTrue );
		static void copyPBProcedureData( DcmSequenceOfItems & procedureSeq, DcmItem & item, OFBool replaceOldValues = OFTrue );

		/**
		 * Copies supported study and patient demography tags from the provided dataset to the document dataset.
		 *
		 * @param dataset     to get the study and patient demography data from
		 * @param replaceOld  true indicates to replace existing data if any (default: OFTrue)
		 *
		 * @return true if and only if study and patient demography data was successfully added to the document dataset
		 */
		virtual bool copySupportedTags( DcmDataset * dataset, OFBool replaceOld = OFTrue );

};

#endif								  // IM_PBR_DOCUMENT_H_INCLUDED
