/*
 * Component: sr
 *
 * Description:
 *  Header file containing the definition of IMSRDocument class.
 *  Also contains the definition of ImageListElem and ImageList classes.
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef IMSRDOC_H_INCLUDED
#define IMSRDOC_H_INCLUDED

#include <iostream>				  /* for class ostream */

#include <dcmtk/dcmsr/dsrdoc.h>

#include "IMDocument.h"
#include "CodeTable.h"
#include "EventList.h"
#include "LogList.h"
#include "PBItem.h"
#include "dtconv.h"

/**
 * Report/Addendum status ID
 */
enum E_PBStatusID {
	PBStatus_CREATED = 50, //!< PBStatus_CREATED
	PBStatus_RECORDED = 60, //!< PBStatus_RECORDED
	PBStatus_TRANSCRIBED = 80,//!< PBStatus_TRANSCRIBED
	PBStatus_APPROVED = 100
//!< PBStatus_APPROVED
};

/**
 * Report Type (Report/Addendum)
 */
enum E_IMSRDocType {
	IMSR_Invalid = 0,//!< IMSR_Invalid
	IMSR_Report = 1, //!< IMSR_Report
	IMSR_Addendum = 2
//!< IMSR_Addendum
};

/**
 * An element in image List
 */
class ImageListElem {
	private:

		char m_id[ 256 ];
		int m_counter;

	public:

		ImageListElem( char * id ) {
			snprintf( m_id, 256, "%s", id );
			m_counter = 1;
		}

		~ImageListElem() {
		}

		inline const char * getId() const {
			return m_id;
		}

		inline int getCounter() const {
			return m_counter;
		}
		inline int incrementCounter() {
			return ++m_counter;
		}

		inline OFBool isIDEquals( const char * id ) const {
			return strcmp( id, m_id ) == 0;
		}

};

/**
 * ImageList is a helper class for dumping figure (Key images and annotations).
 *
 * Since one image can be attached to a Report (and set of reports) more than once
 * with different annotations and or presentation settings (window/level, rotation, ...),
 * we need to distinguish these references.
 *
 * Each image attached to a report will be added to this list with a counter,
 * and the counter will be incremented for every new reference to the image.
 */
class ImageList {
	private:
		OFList< ImageListElem * > m_list;

	public:
		/**
		 * Constructs a new empty <code>ImageList</code>
		 */
		ImageList();

		/**
		 * Destructs this <code>ImageList</code>
		 */
		~ImageList();

		/**
		 * Gets current counter for an image reference.
		 *
		 * If the image reference is not in the list, it will be added to the list
		 * with counter = 1.
		 * If the image reference is in the list, the counter of the image reference will
		 * be incremented.
		 *
		 * @param id  image reference (SOP Instance UID of the image)
		 *
		 * @return the new value of the counter
		 */
		int getImageCounter( char * id );

		/**
		 * Gets the read-only image list
		 * @return read-only image list
		 */
		const OFList< ImageListElem * > & getList() const {
			return m_list;
		}
};

/**
 * Manipulates Structured Report documents created by eRAD software.
 */
class IMSRDocument : public DSRDocument, public IMDocument {
	private:

		/** Code manager used for creating CodedEntryValue objects */
		CodeManager * m_codeManager;

		/** SR Document type (Report/Addendum) */
		E_IMSRDocType m_docType;

		/** Indicates whether this SR document can be modified or not */
		OFBool m_canModify;

	public:

		/**
		 * Constructor for creating a new empty IMSRDocument object
		 *
		 * @param codeManager   code manager used for creating CodedEntryValue objects
		 */
		IMSRDocument( CodeManager * codeManager );

		/**
		 * Constructs a new <code>IMSRDocument</code> for an existing SR document.
		 *
		 * @param filename      File name of the existing SR document
		 * @param codeManager   code manager used for creating CodedEntryValue objects
		 */
		IMSRDocument( const char * filename, CodeManager * codeManager );

		/**
		 * Destructs this <code>IMSRDocument</code>
		 */
		virtual ~IMSRDocument();

		/**
		 * Gets identifier of document type
		 * @return  <code>IMDocSR</code>
		 */
		virtual E_IMDocType getIdent() const {
			return IMDocSR;
		}

		/**
		 * Loads the Structured Report document file
		 *
		 * @param timezone  convert date/times in the dataset to this timezone
		 *                  (NULL: don't convert; "": convert to server local timezone)
		 *
		 * @throw IMException if an error occurs
		 */
		virtual void load( const char * timezone = NULL );

//
//		void testdump();
//
		/**
		 * Saves the Structured Report document into file.
		 *
		 * @param filename  Save the document on this filename. If missing or NULL
		 *                  save on the filename the document was loaded or created.
		 *
		 * @throw IMException if an error occurs
		 */
		virtual void save( const char * filename = NULL );

		/**
		 * Creates new Structured Report document for the given study.
		 * Creates SOP Instance UID and file name for the new document.
		 * Also creates Document Tree structure
		 *
		 * @param studyIUID          Study Instance UID of the study the document will belong to
		 * @param docType            Report|Addendum
		 * @param demogrpahyDataset  if provided, demography data will be copied from this dataset
		 *
		 * @throws IMException if an error occurs
		 */
		void createNew( const char * studyIUID, E_IMSRDocType docType, DcmDataset * demographyDataset = NULL );

		/**
		 * Modifies element in results item.
		 * Since impression text is stored in the content tree, its value will be modified
		 * in the content tree instead of in the results item
		 *
		 * @param elem        identifies the element to change and contains new element value
		 * @param createOnly  if OFTrue then existing element will not be modified
		 *
		 * @return <UL><LI>OFTrue  value of element has been changed</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyResultsElement( PBItem & pbitem, DcmElement * elem, OFBool createOnly = OFFalse );

		/**
		 * Modifies element in interpretation item.
		 * Since observation text and interpretation author are stored in the content tree,
		 * their value will be modified in the content tree instead of in the results item.
		 * Changing Interpretation Recorded Date/Time will take effect by changing Content Date/Time.
		 *
		 * @param elem        identifies the element to change and contains new element value
		 * @param createOnly  if OFTrue then existing element will not be modified
		 *
		 * @return <UL><LI>OFTrue  value of element has been changed</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyInterpretationElement( PBItem & pbitem, DcmElement * elem, OFBool createOnly = OFFalse );

		/**
		 * Modifies Interpretation Status ID element in interpretation item.
		 * It will Complete and Verify Document if new Status ID is 'APPROVED' ('Final')
		 *
		 * @param statusID          contains the new value for the status ID
		 * @param approverSequence  contains Verification information (approver and approval date/time).
		 *                          Should be presented, if statusID is 'APPROVED'.
		 * @param                   return parameter set to OFTrue if the report object was modified.
		 *
		 * @return the new enumerated value of status ID
		 *
		 * @throw IMException if an error occurs
		 */
		E_PBStatusID changeStatus( DcmElement * statusID, DcmSequenceOfItems * approverSequence, OFBool & isChanged );


		/**
		 * Will set the Interpretation Status ID to TRANSCRIBED.
		 * Will also reset the completion flag and removes approver sequence.
		 *
		 * @return true if the SR object could be successfully unfinalized
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool unfinal();

		/**
		 * Checks if the SR document contains the interpretation for given UID or id
		 * Both UID and id might be NULL
		 *
		 * @param uid     (PB) Interpretation SOP Instance UID to search for
		 * @param id      Interpretation ID to search for
		 *
		 * @return <UL><LI>OFTrue  If the document contains the interpretation for the uid or id</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 */
		OFBool containsInterpretation( const char * uid, const char * id );

		/**
		 * Checks if the SR document contains the results for given uid or id
		 * Both uid and id might be NULL
		 *
		 * @param uid     (PB) Results SOP Instance UID to search for
		 * @param id      Results ID to search for
		 *
		 * @return <UL><LI>OFTrue  If the document contains the result for the uid or id</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 */
		OFBool containsResults( const char * uid, const char * id );

		/**
		 * Checks if the given uid equals to the SOP Instance UID of the SR document
		 *
		 * @param uid     SR SOP Instance UID to check
		 *
		 * @return OFTrue  if and only if the SOP Instance UID of the SR document is the same as uid
		 */
		OFBool isUIDEquals( const char * uid );

		/**
		 * Modify demography element values in the SR document by getting
		 * new element values from an item.
		 *
		 * The method will iterate through the elements in item, and will apply changes
		 * in elements of SR document corresponding to the current element in the item.
		 *
		 * Only Demography elements will be changed.
		 *
		 * @param item     contains elements with new values to apply modification
		 */
		void modifyDemography( DcmItem * item );

		inline E_IMSRDocType getIMSRDocumentType() {
			return m_docType;
		}

		/**
		 * Adds a new predecessor document to the Predecessor Document Sequence of the SR Document
		 *
		 * @param studyIUID   Study Instance UID of the predecessor document
		 * @param seriesIUID  Series Instance UID of the predecessor document
		 * @param sopIUID     SOP Instance UID of the predecessor document
		 * @param sopClUID    SOP Class UID of the predecessor document
		 *
		 * @throw IMException if an error occurs
		 */
		void addPredecessorDocument( const char * studyIUID, const char * seriesIUID, const char * sopIUID, const char * sopClUID );

		/**
		 * Dumps (prints) the content of the SR document.
		 *
		 * @param output             the file or device to dump the content of interpretation item on
		 * @param intNum             index of the SR Document
		 * @param imageList          list of attached images
		 *
		 * @throw IMException if an error occurs
		 */
		void dump( FILE * output, int intNum, ImageList * imageList, time_t pbdLastModified );

		/**
		 * Retrieves Report Status (in numeric format)
		 *
		 * @return report status in numeric format
		 */
		int getPBStatus();

		/**
		 * Retrieves Interpretation ID belonging to this document
		 *
		 * @param intID   output parameter for Interpretation ID
		 */
		void getInterpretationID( const char *& intID );

		/**
		 * Retrieves Interpretation Instance UID belonging to this document
		 *
		 * @param intUID   output parameter for Interpretation Instance UID
		 */
		void getInterpretationUID( const char *& intUID );

		/**
		 * Retrieves Results ID belonging to this document
		 *
		 * @param resID   output parameter for Results ID
		 */
		void getResultsID( const char *& resID );

		/**
		 * Retrieves Results Instance UID belonging to this document
		 *
		 * @param resUID   output parameter for Results Instance UID
		 */
		void getResultsUID( const char *& resUID );

		/**
		 * Retrieves Referenced Dictation Instance UID belonging to this document
		 *
		 * @param resUID   output parameter for Referenced Dictation Instance UID
		 */
		void getReferencedDictationUID( const char *& dictUID );

		/**
		 * Completes and verifies the SR document.
		 * The content of the approver sequence will be added to Verifying Observer sequence
		 *
		 * @param approverSequence  Sequence containing Verification information
		 *                         (verifier observer name and verification date/time)
		 *
		 * @throw IMException if an error occurs
		 */
		void completeAndVerifyDocument( DcmSequenceOfItems * approverSequence );

		/**
		 * Updates transcriptionist (data enterer participant) information in the report object.
		 * The item from the participantSequence argument will be copied to the Participant Sequence of the object.
		 * The item will only be copied if the Participant Type is 'ENT'. Pre-existing item in the Participant Sequence
		 * of the object might be overwritten.
		 * The participantSequence argument intended to containing a single item holding participant information in a format
		 * conforming to the Participant Sequence content as defined in DICOM PS 3.3 C.17.2 SR Document General Module.
		 *
		 * @param participantSequence   Transcriptionist data to be added to the report object
		 * @param createOnly            OFTrue indicates to do not add transcriptionist if one already exists
		 *
		 * @return OFTrue if and only if transcriptionist information was added to the report object
		 */
		OFBool updateTranscriptionist( DcmSequenceOfItems * participantSequence, OFBool createOnly );

		/**
		 * Updates author observer information in the report object.
		 * The item from the authorSequence argument will be copied to the Author Observer Sequence of the object.
		 * Pre-existing item in the Author Observer Sequence of the object might be overwritten.
		 * The authorSequence argument intended to containing a single item holding author information in a format
		 * conforming to the Author Observer Sequence content as defined in DICOM PS 3.3 C.17.2 SR Document General Module.
		 *
		 * @param authorSequence   Author data to be added to the report object
		 * @param createOnly       OFTrue indicates to do not add author if one already exists
		 *
		 * @return OFTrue if and only if author information was added to the report object
		 */
		OFBool updateAuthor( DcmSequenceOfItems * authorSequence, OFBool createOnly );

		/**
		 * Modifies the content of this SR Document by applying changes in
		 * interpretation related attributes. Such attributes are:
		 *  - Observation text
		 *  - Author, Transcriptionist information (Name, Date/Time)
		 *  - Verifying (Approving) information (Name(s), Date/Time(s))
		 *  - Report Status
		 *
		 * @param item       contains new values for interpretation related attributes
		 * @param eventList  put generated events to this list (triggers sending HL7 and DICOM messages)
		 * @param logList    add log entry to log list for logging incoming report
		 * @param isNew      is this a new SR document or not
		 *
		 * @throw IMException if an error occurs
		 */
		void processInterpretationItem( DcmItem * item, EventList & eventList, LogList & logList, OFBool isNew );

		/**
		 * Modifies the content of this SR Document by applying changes in
		 * results related attributes. Such attribute is:
		 *  - Impression text
		 *
		 * @param item       contains new values for results related attributes
		 * @param eventList  put generated events to this list (triggers sending HL7 and DICOM messages)
		 * @param isNew      is this a new SR document or not
		 *
		 * @throw IMException if an error occurs
		 */
		void processResultsItem( DcmItem * item, EventList & eventList, OFBool isNew );

		/**
		 * Modifies the content of this SR Document by applying changes stored
		 * in a dataset.
		 *
		 * @param dataset           contains new elements and/or new values for this SR Document
		 * @param eventList         put generated events to this list (triggers sending HL7 and DICOM messages)
		 * @param logList           add log entry to log list for logging incoming report
		 * @param isNew             is this a new SR document or not
		 * @param updateDemography  true indicates to also update demography data taken from the dataset
		 *
		 * @throw IMException if an error occurs
		 */
		void processModificationDataset( pbtz::OffsetAdjustedDataset &ofDataset, EventList & eventList, LogList & logList, OFBool isNew, OFBool updateDemography = OFTrue );

		// Merge Studies with reports

		inline DSRSOPInstanceReferenceList * getPredecessorDocumentSequence() {
			return &PredecessorDocuments;
		}

		inline void setStudyInstanceUID( const char * uid ) {
			if ( m_studyIUID != NULL )
				delete m_studyIUID;
			m_studyIUID = new char[ strlen( uid ) + 1 ];

			strcpy( m_studyIUID, uid );
		}

		inline void setSOPInstanceUID( const char * uid ) {
			if ( m_sopIUID != NULL )
				delete m_sopIUID;
			m_sopIUID = new char[ strlen( uid ) + 1 ];

			strcpy( m_sopIUID, uid );
		}

		inline DcmUniqueIdentifier *getSOPInstanceUIDElement() {
			return &SOPInstanceUID;
		}

		inline DcmUniqueIdentifier *getStudyInstanceUIDElement() {
			return &StudyInstanceUID;
		}

		inline void setCanModify( OFBool canModify ) {
			m_canModify = canModify;
		}

		/**
		 * Adds attached image objects to imageList.
		 *
		 * @param imageList   add attached image objects to this list
		 *
		 * @throw IMException if an error occurs
		 */
		void listFigureObjects( ImageList * imageList );

		/**
		 * Degrade Report SR to Addendum
		 *
		 * Will only be used in case merging studies will attach reports from
		 * non primary studies as addendum to report of primary study.
		 */
		void degradeToAddendum();

		/**
		 * Makes Report SR from Addendum
		 *
		 * Will only be used in merge SRs
		 */
		void makeReportFromAddendum();

		/**
		 * Retrieves the impression which is the value of (121073^DCM^Impression TEXT)
		 * in the (121072^DCM^Impressions CONTAINER) tree node.
		 *
		 * @return the value of the impression
		 */
		OFString getImpressionsText();

		/**
		 * Retrieves the findings (observation) which is the value of (121071^DCM^Finding TEXT)
		 * in the (121070^DCM^Findings CONTAINER) tree node if the SR contains a report or
		 * in the (121078^DCM^Addendum CONTAINER) tree node if the SR contains an addendum.
		 *
		 * @return the value of the impression
		 */
		OFString getFindingsText();

		/**
		 * Retrieves HTML impression which is the value of (38100^IMPB^Html Impression TEXT)
		 * in the (121072^DCM^Impressions CONTAINER) tree node.
		 *
		 * @return the value of HTML impression
		 */
		OFString getHtmlImpressionsText();

		/**
		 * Retrieves HTML findings (observation) which is the value of (38110^IMPB^Html Finding TEXT)
		 * in the (121070^DCM^Findings CONTAINER) tree node if the SR contains a report or
		 * in the (121078^DCM^Addendum CONTAINER) tree node if the SR contains an addendum.
		 *
		 * @return the value of the impression
		 */
		OFString getHtmlFindingsText();

		/**
		 * Retrieves the findings (observation) which is the value of (121071^DCM^Finding TEXT)
		 * in the (121070^DCM^Findings CONTAINER) tree node if the SR contains a report or
		 * in the (121078^DCM^Addendum CONTAINER) tree node if the SR contains an addendum.
		 *
		 * @return the value of the impression
		 */
		OFString getFindingsNote();

		/**
		 * Retrieves the findings (observation) which is the value of (121071^DCM^Finding TEXT)
		 * in the (121070^DCM^Findings CONTAINER) tree node if the SR contains a report or
		 * in the (121078^DCM^Addendum CONTAINER) tree node if the SR contains an addendum.
		 *
		 * @return the value of the impression
		 */
		OFString getFindingsNormality();


		void getAuthorInfo( OFString & authorID, OFString & authorName );
		void getTranscriptionInfo( OFString & transcriptionistID, OFString & transcriptionistName, OFString & transcriptionDate, OFString & transcriptionTime );
		void getApproverInfo( OFString & approverID, OFString & approverName, OFString & approveDate, OFString & approveTime );


		/**
		 * Adds a new verifier (approver) to the Verifying Observer sequence
		 *
		 * @param observerName   Verifying Observer Name (Approver)
		 * @param observerCode   Verifying Observer Code (Not used)
		 * @param organization   Organization of the Observer (??/Not used)
		 * @param datetime       Date Time of Verification (Approval Date/Time)
		 *
		 * @return <UL><LI>EC_Normal    if verifier has been successfully added</LI>
		 *             <LI>EC_IllegalCall, EC_MemoryExhausted otherwise</LI></UL>
		 */
		OFCondition verifyDocument( const OFString & observerName, const DSRCodedEntryValue & observerCode,
				const OFString & organization, const OFString & datetime );


		/**
		 * Copies supported study and patient demography tags from the provided dataset to the document dataset.
		 *
		 * @param dataset     to get the study and patient demography data from
		 * @param replaceOld  true indicates to replace existing data if any (default: OFTrue)
		 *
		 * @return true if and only if study and patient demography data was successfully added to the document dataset
		 */
		virtual bool copySupportedTags( DcmDataset * dataset, OFBool replaceOld = OFTrue );

	private:
		OFString m_PBStatus;

		DcmSequenceOfItems * m_PBResultsSequence;
		DcmSequenceOfItems * m_PBInterpretationSequence;

		DcmSequenceOfItems * m_authorObserverSequence;
		DcmSequenceOfItems * m_participantSequence;

		/**
		 * Retrieves a top level (right under root) node.
		 *
		 * @param conceptName  identifies the container to be retrieved.
		 * @param baseNode     root node of the subtree to be searched (NULL means DocumentRoot)
		 *
		 * @return  pointer to the Node
		 *
		 * @throw IM_ASSERT exception if the root node was not found
		 */
		size_t findNode( DSRCodedEntryValue & conceptName, size_t baseNode = 0 );

		/**
		 * Retrieves a top level (right under root) node.
		 *
		 * @param conceptName        identifies the container to be retrieved.
		 * @param baseNode           root node of the subtree to be searched (NULL means DocumentRoot)
		 * @param relationshipType   type of relationship to parent node
		 * @param valueType          value type of the node
		 *
		 * @return  pointer to the Node
		 *
		 * @throw IM_ASSERT exception if the root node was not found
		 */
		size_t findNode( DSRCodedEntryValue & conceptName, const E_RelationshipType relationshipType, const E_ValueType valueType,
				size_t baseNode = 0 );

		size_t findOrCreateNode( DSRCodedEntryValue & conceptName, const E_RelationshipType relationshipType,
				const E_ValueType valueType, size_t baseNode = 0, size_t prevNode = 0 );

		/**
		 * Finds and retrieves a Text Tree Node with its Container Tree Node from the content tree.
		 * This method is used for retrieving impression and observation (findings, addendum) texts.
		 *
		 * @param containerName  Identifies the container tree node
		 * @param textName       Identifies the text tree node
		 * @param containerNode  Output parameter for the container node
		 * @param textNode       Output parameter for the text node
		 */
		void getTextWithContainer( DSRCodedEntryValue & containerName, DSRCodedEntryValue & textName, size_t & containerNode,
				size_t & textNode );

		/**
		 * Finds and retrieves a Text Tree Node with its Container Tree Node from the content tree.
		 * The existence of the text and container tree nodes will be checked.
		 * This method is used for retrieving impression and observation (findings, addendum) texts.
		 *
		 * @param containerName  Identifies the container tree node
		 * @param textName       Identifies the text tree node
		 * @param containerNode  Output parameter for the container node
		 * @param textNode       Output parameter for the text node
		 *
		 * @throw IMException if either the container node or the text node is missing.
		 */
		void getAndCheckTextWithContainer( DSRCodedEntryValue & containerName, DSRCodedEntryValue & textName, size_t & containerNode,
				size_t & textNode );

		/**
		 * Sets string value of the current tree node in the content tree
		 *
		 * @param strValue   the new value of the tree node
		 *
		 * @return <UL><LI>OFTrue   if the value of the current node has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool setCurrentStringValue( const char * strValue );

		/**
		 * Sets coded entry value of the current tree node in the content tree
		 *
		 * @param codedEntryValue   the new value of the tree node
		 *
		 * @return <UL><LI>OFTrue   if the value of the current node has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool setCurrentCodeValue( const DSRCodedEntryValue & codedEntryValue );

		/**
		 * Remove Observation Context encoded in the content tree of legacy eRAD report objects.
		 */
		void removeLegacyObsContext();

		/**
		 * Modifies the content of (121072^DCM^Impressions CONTAINER) tree node.
		 * This container contains the (121073^DCM^Impression TEXT) tree node,
		 * which is the impression of the report.
		 *
		 * Only the string value stored in the Impression node can be changed.
		 *
		 * @param key        Identifies the field to be changed
		 *                     DCM_Impressions for Impression Text
		 *                     DCM_PBHtmlImpressions for HTML Impressions Text
		 * @param strValue   The new value of the field
		 *
		 * @return <UL><LI>OFTrue   if the value of the impression has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyImpressionsContainer( PBItem & pbitem, DcmTagKey & key, const char * strImpressions );

		/**
		 * Modifies the content of (121072^DCM^Findings CONTAINER) or (121078^DCM^Addendum) tree node.
		 * This container contains the (121071^DCM^Finding TEXT) tree node,
		 * which is the observation of the report or addendum.
		 * This container also has an observation context (121008^DCM^Person Observer's Name PNAME),
		 * which stands for the interpretation author.
		 *
		 * Only the value of Finding or Person Observer's Name can be changed
		 *
		 * @param key        Identifies the field to be changed
		 *                     DCM_InterpretationAuthor for Person Observer's Name
		 *                     DCM_InterpretationText for Finding
		 *                     DCM_PBHtmlInterpretationText for HTML Interpretation Text
		 *                     DCM_ResultsComments for Radiologist's Note
		 *                     DCM_PBNormalIndication for Normal Indication Code
		 * @param strValue   The new value of the field
		 * @param createOnly OFTrue means do not modify existing values, only create it if not yet exists
		 *
		 * @return <UL><LI>OFTrue   if the value of the Findings an/or Person Observer's Name has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyFindingsContainer( PBItem & pbitem, DcmTag & key, const char * strValue, OFBool createOnly = OFFalse );

		/**
		 * Sets string value of the current tree node in the content tree. The new value of the
		 * tree node will be read from a file
		 *
		 * @param filename   name of the file containing the new value for the current tree node
		 *
		 * @return <UL><LI>OFTrue   if the value of the current node has been changed</LI>
		 *             <LI>OFFalse  otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyStringFromFile( const char * filename );

		void checkAndFixIDs( DcmItem & item, DcmTagKey idTag, const char * uidPrefix, OFString & origSop, OFString & sop,
				OFString & origId, OFString & id );
		void addInstanceCreationAttributes( DcmItem & item );

		/**
		 * Finds and retrieves the string value of an element from an item
		 *
		 * @param src    the item to search for the element
		 * @param key    identifies the element to find
		 * @param value  output parameter for the element value
		 *
		 * @return <UL><LI>OFTrue  if the element and its value has been successfully retrieved</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool getElementValue( DcmItem * src, DcmTagKey key, const char *& value );

		void checkTypeID( const char * typeID );
		E_PBStatusID checkAndCopyStatusID( DcmItem * item, const char * statusID );

		/**
		 * Checks if the given tag is a demography tag or not
		 *
		 * @param tag     the tag to check
		 *
		 * @return OFTrue  if and only if the tag is NOT a demography tag
		 */
		OFBool isNotDemographyTag( DcmTagKey & tag );

		/**
		 * Reads SR content tree (and SR related tags) from dataset
		 * Also reads results and interpretation sequences
		 *
		 * @param dataset  contains the SR content tree
		 * @param flags    flags applied while reading (see SRDocument::read())
		 *
		 * @return EC_Normal  if content has been successfully read or error code
		 * @see SRDocument::read()
		 *
		 * @throw IMException if an error occurs
		 */
		OFCondition read( DcmItem & dataset, const size_t flags );

		/**
		 * Writes SR content tree (and SR related tags) into dataset
		 * Also writes results and interpretation sequences
		 *
		 * @param dataset      write SR content tree into this dataset
		 * @param markedItems  (see SRDocument::write())
		 *
		 * @return EC_Normal  if content has been successfully written or error code
		 * @see SRDocument::write()
		 *
		 * @throw IMException if an error occurs
		 */
		OFCondition write( DcmItem & dataset, DcmStack * markedItems );

		/**
		 * Modifies element in the dataset.
		 *
		 * @param elem        identifies the element to change and contains new element value
		 * @param createOnly  if OFTrue then existing element will not be modified
		 *
		 * @return <UL><LI>OFTrue  value of element has been changed</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyElement( DcmElement * elem, OFBool createOnly = OFFalse );

		/**
		 * Modifies element in the first item of the sequence.
		 *
		 * @param sequence    modify element in the first item of this sequence
		 * @param elem        identifies the element to change and contains new element value
		 * @param createOnly  if OFTrue then existing element will not be modified
		 *
		 * @return <UL><LI>OFTrue  value of element has been changed</LI>
		 *             <LI>OFFalse otherwise</LI></UL>
		 *
		 * @throw IMException if an error occurs
		 */
		OFBool modifyElementInSequence( DcmSequenceOfItems * sequence, DcmElement * elem, OFBool createOnly );

		/**
		 * Modifies the (string) value of an element by copying the value of an another element
		 *
		 * @param elem     the element to modify
		 * @param modElem  copy the value of this element into elem
		 *
		 * @return OFTrue  if and only if the value of elem has been changed
		 */
		OFBool modifyElemValue( DcmElement & elem, DcmElement * modElem );

		//	DcmItem * getOrCreatePredecessorItemForStudy(const char * studyIUID);
		//	DcmItem * getOrCreatePredecessorItemForSeries(DcmSequenceOfItems * seq, const char * seriesIUID);


		/**
		 * Dumps (prints) the content of the results sequence item from the SR document.
		 *
		 * @param output  the file or device to dump the content of results item on
		 *
		 * @throw IMException if an error occurs
		 */
		void dumpResults( FILE * output );


		/**
		 * Dumps author (Author Observation) information.
		 *
		 * @param output             the file or device to dump the author information on
		 * @param intNum             index of the SR Document
		 */
		void dumpAuthor( FILE * output, int intNum );

		/**
		 * Dumps transcription (data enterer Participant) information.
		 *
		 * @param output             the file or device to dump the transcription information on
		 * @param intNum             index of the SR Document
		 */
		void dumpTranscriptionist( FILE * output, int intNum );


		/**
		 * Dumps (prints) the content of the interpretation sequence item from the SR document.
		 *
		 * @param output             the file or device to dump the content of interpretation item on
		 * @param intNum             index of the SR Document
		 * @param pbdLastModified    date/time of the PbD file if there is a dictation recorded for this SR Document
		 *
		 * @throw IMException if an error occurs
		 */
		void dumpInterpretation( FILE * output, int intNum, time_t pdbLastModified );

		/**
		 * Dumps (prints) approver (verifying observer) information from the SR document.
		 *
		 * @param output             the file or device to dump the content of interpretation item on
		 * @param intNum             index of the SR Document
		 *
		 * @throw IMException if an error occurs
		 */
		void dumpApprover( FILE * output, int intNum );

		/**
		 * Dumps (prints) figure (attached images and annotations) information from the SR document.
		 *
		 * @param output             the file or device to dump the content of interpretation item on
		 * @param intNum             index of the SR Document
		 * @param imageList          list of attached images
		 *
		 * @throw IMException if an error occurs
		 */
		void dumpFigure( FILE * output, int intNum, ImageList * imageList );

		DSRCodedEntryValue normalityToCEV( const char * normalityStatus );

		// --- declaration copy constructor and assignment operator

		IMSRDocument( const IMSRDocument & );
		IMSRDocument & operator=( const IMSRDocument & );
};

#endif

