/*
 * Component: hl7
 *
 * Description:
 *   Header file for class definitions to load order/study data to be added to outgoing HL7 messages
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __PB_ORDER_DATA_FILLER_H_INCLUDED__
#define __PB_ORDER_DATA_FILLER_H_INCLUDED__

#include <map>
#include <dcmtk/dcmdata/dctk.h>

#include "IMOrderManager.h"

#define FAKE_PROCEDURE_STEP_ID "FAKE_PROCEDURE_STEP_ID_FOR_PROCESSING"

class OrderDataFiller;


/**
 * Wrapper class around study derivation information entries
 */
class DerivationInfoEntry {

	private:
		/** Instance UID of the original study */
		OFString m_origStudy;

		/** Instance UID of the derived study */
		OFString m_derivedStudy;

		/** Derivation operation performed to create the derived study from the original study */
		OFString m_oper;

		/** Folder the study copied/linked to */
		OFString m_folder;

	public:

		/**
		 * Constructor to create a new DerivationInfoEntry object
		 *
		 * @param origStudy      Instance UID of the original study
		 * @param derivedStudy   Instance UID of the derived study
		 * @param oper           Derivation operation performed to create the derived study from the original study
		 * @param folder         Folder the study copied/linked to
		 */
		DerivationInfoEntry( const OFString & origStudy, const OFString & derivedStudy, const OFString & oper, const OFString & folder );

		/**
		 * Copy constructor to create a new DerivationInfoEntry object by copying an existing one
		 *
		 * @param orig   Original DerivationInfoEntry object to copy
		 */
		DerivationInfoEntry(const DerivationInfoEntry & orig);

		/**
		 * Destructor to destroy this DerivationInfoEntry object
		 */
		~DerivationInfoEntry();

		/**
		 * @return  Instance UID of the original study
		 */
		inline const OFString & getOrigStudy() const {
			return m_origStudy;
		}

		/**
		 * @return Instance UID of the derived study
		 */
		inline const OFString & getDerivedStudy() const {
			return m_derivedStudy;
		}

		/**
		 * @return Derivation operation performed to create the derived study from the original study
		 */
		inline const OFString & getOper() const {
			return m_oper;
		}

		/**
		 * @return Folder the study copied/linked to
		 */
		inline const OFString & getFolder() const {
			return m_folder;
		}

		/**
		 * Assignment operator. Copies the content of the argument DerivationInfoEntry object to this object
		 *
		 * @param orig   DerivationInfoEntry object to be copied
		 *
		 * @return reference to this DerivationInfoEntry object
		 */
		DerivationInfoEntry & operator= ( const DerivationInfoEntry & orig );
};

/**
 * Wrapper class around additional study information to be mapped to outgoing HL7 messages.
 *
 * Collects and provides the following study information
 * - Derivation information for derived studies
 * - Archive device for retrieved studies
 */
class AdditionalStudyInfo {
	private:
		/** Derivation information for studies derived from other studies */
		OFList<DerivationInfoEntry> m_derivationInfo;

		/** ID of the device the study was retrieved from (if retrieved at all) */
		OFString m_retrievedFrom;

		/**
		 * Reads derivation info file for studies merged/split
		 * @param study
		 */
		void checkAndReadDerivationInfo( const OFString & study );

	public:

		/**
		 * Constructor to create a new AdditionalStudyInfo object
		 */
		AdditionalStudyInfo();

		/**
		 * Destructor destroying this AdditionalStudyInfo object
		 */
		~AdditionalStudyInfo();


		/**
		 * Collects additional study information for the provided study
		 *
		 * @param study  collect additional study information for this study
		 */
		void collect( const OFString & study );

		/**
		 * @return Derivation information for studies derived from other studies
		 */
		inline const OFList<DerivationInfoEntry> & getDerivationInfo() const {
			return m_derivationInfo;
		}

		/**
		 * @return ID of the device the study was retrieved from (if retrieved at all)
		 */
		inline const OFString & getRetrievedFrom() const {
			return m_retrievedFrom;
		}
};

/**
 * Wrapper class around an order (scheduled step).
 *
 * Consist of a pair of request item and scheduled step item.
 */
class ProcedureItems {
	friend class OrderDataFiller;

private:

	/** Request Item to hold request level data of the order */
	DcmItem * m_requestItem;

	/** Scheduled Step Item to hold scheduled step level data of an order */
	DcmItem * m_scheduledStepItem;

protected:

	/**
	 * Constructor for creating a new ProcedureItems object
	 *
	 * @param requestItem          Request Item to hold request level data of the order
	 * @param scheduledStepItem    Scheduled Step Item to hold scheduled step level data of an order
	 */
	ProcedureItems(DcmItem * requestItem, DcmItem * scheduledStepItem) {
		m_requestItem = requestItem;
		m_scheduledStepItem = scheduledStepItem;
	}
	
	/**
	 * Returns the Request Item
	 *
	 * @return the Request Item
	 */
	inline DcmItem * getRequestItem() const {
		return m_requestItem;
	}

	/**
	 * Returns the Scheduled Step Item
	 *
	 * @return the Scheduled Step Item
	 */
	inline DcmItem * getScheduledStepItem() const {
		return m_scheduledStepItem;
	}

	/**
	 * Assignment operator. Copies the content of the argument ProcedureItems object to this object
	 *
	 * @param orig   ProcedureItems object to be copied
	 *
	 * @return reference to this ProcedureItems object
	 */
	inline ProcedureItems & operator= (const ProcedureItems & orig) {
		if (this != &orig) {
			m_requestItem = orig.m_requestItem;
			m_scheduledStepItem = orig.m_scheduledStepItem;
		}

		return *this;
	}

public:

	/**
	 * Copy constructor.
	 *
	 * @param orig   ProcedureItems object to be copied
	 */
	ProcedureItems(const ProcedureItems & orig) {
		m_requestItem = orig.m_requestItem;
		m_scheduledStepItem = orig.m_scheduledStepItem;
	}
	

	/**
	 * Checks if this ProcedureItems object matches to a procedure step.
	 * The ProcedureItems object matches to the procedure step, if the scheduled
	 * procedure step item contains a ScheduledProcedureStepId element and the value of
	 * the ScheduledProcedureStepId element equals to procedureStepId
	 * 
	 * @param procedureStepId  Value for the Scheduled Procedure Step ID to be checked
	 *
	 * @return true if and only if this ProcedureItems object matches to the procedure step
	 */
	bool isMatchingTo(const char * procedureStepId) {
		if (m_scheduledStepItem == NULL)
			return false;

		const char * stepId = NULL;
		m_scheduledStepItem->findAndGetString(DCM_ScheduledProcedureStepID, stepId);
		return (stepId != NULL) && (strcmp(stepId, procedureStepId) == 0);
	}
};


/**
 * Helper class to fill order data from the PbR file of the study and from the Dcstudy table for outgoing HL7 messages.
 *
 * Intended usage:
 *
 *   <code>
 *   OrderDataFiller odf(study);
 *   if (!initialize()) {
 *   	  // error to be handled
 *   }
 *
 *   OFList<ProcedureItems> & orders = odf.getProcedureSteps();
 *
 *   OFIterator<OFString> * cur = orders.begin();
 *   OFIterator<OFString> * end = orders.end();
 *   while ( cur != end) {
 *   	  ProcedureItems & pItems = *cur;
 *   	  DcmDataset * dataset = odf.getDatasetForProcedureStep(pItems);
 *      // create the HL7 message from the dataset
 *      ...
 *      delete dataset;
 *   }
 *   </code>
 *
 * Date/Time values in datasets returned by OrderDataFiller shall use local time and
 * the dataset shall not contain (0008,0201) TimezoneFromUTCOffset tag.
 */
class OrderDataFiller {
private:

	/** Study Instance UID of the study */
	OFString m_study;

	/** Order manager for reading the PbR file of the study */
	IMOrderManager * m_orderManager;

	/** Dataset of the PbR file of the study */
	DcmDataset * m_pbrDataset;

	/** List of procedure steps in the PbR file */
	OFList<ProcedureItems> m_procedureSteps;

	/** Dataset created from the study table content */
	DcmDataset * m_databaseDataset;

	/** Container for objects to be removed from pbr dataset but referred by request/scheduled step items */
	OFList<DcmElement *> m_removedObjects;

	/** Indicates whether this object was created for passing information to the scheduler */
	OFBool m_forScheduler;

	/** User defined custom fields - filled from the study table entry */
	std::map<OFString, OFString> m_customFields;
	
	/** Additional information for derived and retrieved studies */
	AdditionalStudyInfo m_additionalStudyInfo;

	/** true indicates to only read values from the database (default is false)*/
	OFBool m_databaseOnly;

	/**
	 * Fills database dataset from study table.
	 *
	 * @returns true if and only if database dataset was successfully filled.
	 */
	bool fillDatabaseDataset();

public:

	/**
	 * Constructor for creating a new OrderDataFiller object.
	 *
	 * @param study   Study Instance UID of the study 
	 */
	OrderDataFiller(const char * study, OFBool forScheduler = OFFalse);

	/**
	 * Destructor for cleaning up OrderDataFiller object.
	 */
	~OrderDataFiller();


	/**
	 * Initialize the OrderDataFiller object. Will read order data from PbR file and study table.
	 *
	 * @return true if and only if the OrderDataFiller object was successfully initialized
	 */
	bool initialize();

	/**
	 * Prints order items (Request/Scheduled Procedure Step item pair) to standard output.
	 * This method is for debug purpose.
	 *
	 * @param items   order items to be printed
	 */
	void printItems(const ProcedureItems & items);
	

	/**
	 * Get dataset filled with values of order items (Request/Scheduled Procedure Step item pair).
	 * Values will be taken from:
	 *   - Scheduled Procedure Step item
	 *   - Request item
	 *   - PbR dataset root
	 *   - study table
	 *
	 * Date/Time values in the dataset shall use be in local time zone and the dataset
    * shall not contain (0008,0201) TimezoneFromUTCOffset tag.
	 *
	 * @param items   order items to be printed
	 *
	 * @return the dataset filled with values of the order
	 */
	DcmDataset * getFilledDataset(const ProcedureItems & items);


	/**
	 * Get dataset filled with values of order items (Request/Scheduled Procedure Step item pair).
	 * Values will be taken from:
	 *   - Scheduled Procedure Step item (the first item if any)
	 *   - Request item (the first item if any)
	 *   - PbR dataset root
	 *   - study table
	 *
	 * In Database Only mode it will return values from the study table only
	 *
	 * Date/Time values in the dataset shall use be in local time zone and the dataset
    * shall not contain (0008,0201) TimezoneFromUTCOffset tag.
	 *
	 * @param items   order items to be printed
	 *
	 * @return the dataset filled with values of the order
	 */
	DcmDataset * getFilledDataset();


	/**
	 * Return the list of orders (Request/Scheduled Procedure Step item pairs) for the study.
	 *
	 * @return the list of orders for the study
	 */
	inline const OFList<ProcedureItems> & getProcedureSteps() {
		return m_procedureSteps;
	}

	/**
	 * @return  Additional information for derived and retrieved studies
	 */
	const AdditionalStudyInfo & getAdditionalStudyInfo() const {
		return m_additionalStudyInfo;
	}

	/**
	 * Sets whether study attributes to be read from the database only (default is false)
	 *
	 * @param databaseOnly  true indicates to read study attributes from the database only
	 */
	inline void setDatabaseOnly( const OFBool & databaseOnly ) {
		m_databaseOnly = databaseOnly;
	}

};

#endif // __PB_ORDER_DATA_FILLER_H_INCLUDED__
