/*
 * SRMap.h
 *
 *  Created on: Aug 11, 2014
 *      Author: tnemeth
 */

#ifndef _IM_SR_FORMATTER_H_
#define _IM_SR_FORMATTER_H_

#include <ostream>

#include <dcmtk/dcmsr/dsrdoc.h>
#include <dcmtk/dcmsr/dsrstrvl.h>
#include <dcmtk/dcmsr/dsrcodtn.h>
#include <dcmtk/dcmsr/dsrcomtn.h>
#include <dcmtk/dcmsr/dsrcontn.h>
#include <dcmtk/dcmsr/dsrdattn.h>
#include <dcmtk/dcmsr/dsrdoctn.h>
#include <dcmtk/dcmsr/dsrdtitn.h>
#include <dcmtk/dcmsr/dsrimgtn.h>
#include <dcmtk/dcmsr/dsrnumtn.h>
#include <dcmtk/dcmsr/dsrpnmtn.h>
#include <dcmtk/dcmsr/dsrreftn.h>
#include <dcmtk/dcmsr/dsrscotn.h>
#include <dcmtk/dcmsr/dsrtcotn.h>
#include <dcmtk/dcmsr/dsrtextn.h>
#include <dcmtk/dcmsr/dsrtimtn.h>
#include <dcmtk/dcmsr/dsruidtn.h>
#include <dcmtk/dcmsr/dsrwavtn.h>


/**
 * Name space providing tools to format the content of Structured Report objects
 */
namespace imsrmapping {


/**
 * Base class providing methods to store formatted content
 */
class Buffer {
	protected:
		/** The buffer the formatted content to be added to */
		OFString m_buffer;

		/** Suffix for the format-dependent configuration files */
		OFString m_configSuffix;

	public:

		/**
		 * Default constructor to create a new empty Buffer object
		 */
		Buffer();

		/**
		 * Destructor to destroy this Buffer object
		 */
		virtual ~Buffer();

		/**
		 * Encodes content string accordingly to the output format and appends the encoded content to the buffer.
		 *
		 * @param str  the content to be appended to the buffer
		 */
		virtual void appendEncoded( const OFString & str ) = 0;

		/**
		 * Appends a single character to the buffer as is.
		 *
		 * @param c  the character to be appended to the buffer
		 */
		inline void append( const char c ) {
			m_buffer.append( 1, c );
		}

		/**
		 * Appends a string to the buffer as is.
		 * The string intended to be already formatted according to the output format.
		 *
		 * @param str  the string to be appended to the buffer
		 */
		inline void append( const char * str ) {
			m_buffer.append( str );
		}

		/**
		 * Appends a string to the buffer as is.
		 * The string intended to be already formatted according to the output format.
		 *
		 * @param str  the string to be appended to the buffer
		 */
		inline void append( const OFString & str ) {
			m_buffer.append( str );
		}

		/**
		 * @return the formatted content
		 */
		inline const char * getString() const {
			return m_buffer.c_str();
		}

		/**
		 * @return suffix for the format-dependent configuration files
		 */
		inline const OFString & getConfigSuffix() {
			return m_configSuffix;
		}

		/**
		 * @return true if and only if the formatted content is empty
		 */
		inline bool isEmpty() {
			return m_buffer.empty();
		}

		/**
		 * @return the length of the formatted content
		 */
		inline size_t getLength() {
			return m_buffer.length();
		}
};


/**
 * Buffer for content formatted as plain text.
 */
class TextBuffer : public Buffer {
	public:
		/**
		 * Default constructor to create a new empty TextBuffer object
		 */
		TextBuffer();

		/**
		 * Destructor to destroy this TextBuffer object
		 */
		virtual ~TextBuffer();

		/**
		 * Encodes content string accordingly to the output format and appends the encoded content to the buffer.
		 * appendEncoded() in TextBuffer appends the string as is.
		 *
		 * @param str  the content to be appended to the buffer
		 */
		virtual void appendEncoded( const OFString & str );
};


/**
 * Buffer for content formatted as HTML code.
 */
class HtmlBuffer : public Buffer {
	public:
		/**
		 * Default constructor to create a new empty HtmlBuffer object
		 */
		HtmlBuffer();

		/**
		 * Destructor to destroy this HtmlBuffer object
		 */
		virtual ~HtmlBuffer();

		/**
		 * Encodes content string accordingly to the output format and appends the encoded content to the buffer.
		 * Content will be HTML encoded.
		 *
		 * @param str  the content to be appended to the buffer
		 */
		virtual void appendEncoded( const OFString & str );
};


/**
 * Describes formatting rules for nodes matching to a specific pattern in the SR content tree,
 *
 * Node patterns include relationship type, value type and concept name.
 *
 * Node patterns are ordered from most specific to least specific, and applied in this order.
 *
 */
class FormatRule {
	private:

		/** Relationship type of the node to its parent node */
		DSRTypes::E_RelationshipType m_relationshipType;

		/** Value type of the node */
		DSRTypes::E_ValueType m_valueType;

		/** Concept name of the node */
		DSRCodedEntryValue m_conceptName;

	protected:
		/** Format string specified for the matching nodes */
		OFString m_format;

	public:

		/**
		 * Default constructor to create a new Format Rule object matching to any node and having no format string specified.
		 */
		FormatRule();
		/**
		 * Destructor to destroy this Format Rule object.
		 */
		~FormatRule();

		/**
		 * Sets relationship type for this rule.
		 *
		 * RT_unknown and RT_invalid mean to match any relationship type
		 *
		 * @param relationshipType   new relationship type to be set for this rule
		 */
		inline void setRelationshipType( const DSRTypes::E_RelationshipType & relationshipType ) {
			if ( relationshipType == DSRTypes::RT_unknown ) {
				m_relationshipType = DSRTypes::RT_invalid;
			}
			else {
				m_relationshipType = relationshipType;
			}
		}

		/**
		 * Sets value type for this rule.
		 *
		 * VT_invalid mean to match any value type
		 *
		 * @param valueType   new value type to be set for this rule
		 */
		inline void setValueType( const DSRTypes::E_ValueType & valueType ) {
			m_valueType = valueType;
		}

		/**
		 * Sets concept name for this rule.
		 *
		 * Empty concept name mean to match any concept name
		 *
		 * @param conceptName   new concept name to be set for this rule
		 */
		inline void setConceptName( const DSRCodedEntryValue & conceptName ) {
			m_conceptName = conceptName;
		}

		/**
		 * @return relationship type set for this rule
		 */
		inline const DSRTypes::E_RelationshipType & getRelationshipType() {
			return m_relationshipType;
		}

		/**
		 * @return value type set for this rule
		 */
		inline const DSRTypes::E_ValueType & getValueType() {
			return m_valueType;
		}

		/**
		 * @return concept name set for this rule
		 */
		inline const DSRCodedEntryValue & getConceptName() {
			return m_conceptName;
		}


		/**
		 * Sets format string for this rule.
		 *
		 * @param format   new format string to be set for this rule
		 */
		inline void setFormat( const OFString & format ) {
			m_format = format;
		}

		/**
		 * @return format string set for this rule
		 */
		inline const OFString & getFormat() const {
			return m_format;
		}

		/**
		 * Checks whether the specified node matches to this rule or not.
		 *
		 * @param node   node to check
		 * @return true if and only if the node matches to this rule.
		 */
		bool isNodeMatching( const DSRDocumentTreeNode & node ) const;


		/**
		 * Compares this rule to the specified rule.
		 *
		 * @param rule    rule to compare
		 * @return
		 *      <0  if this rule is before the argument rule (argument rule is less specific)
		 *       0  if the two rules are the same (has the same relationship type, value type and concept name)
 	 	 *      >0  if this rule is after the argument rule (argument rule is more specific)
		 */
		int compareTo( const FormatRule & rule ) const;


		/**
		 * Prints the rule to the specified output stream in a readable format.
		 *
		 * @param out   output stream to write to
		 */
		void dump( std::ostream & out );
};




/**
 * Formats the content of an SR object according to the configured rules
 */
class Formatter {
	private:

		/** Content tree of the SR object */
		DSRDocumentTree * m_tree;

		/** true indicate to print verbose messages to standard output */
		bool m_verbose;

	protected:

		/** Formatting rules read from configuration files */
		OFList< FormatRule * > m_formatRules;

		/** Default (empty) format for nodes not matching to any formatting rules */
		OFString m_defaultFormat;

		/** Header formats - allows to use different headers at different levels */
		OFMap< size_t, OFString > m_headers;

		/** Default header format */
		OFString m_defaultHeaderFormat;

		/** Current level in the tree */
		size_t m_level;

		/** Buffer to write to formatted content to */
		Buffer & m_buffer;

		/**
		 * Formats the content of the specified node by replacing fields in the format string with values taken from the node,
		 * and appends the formatted content the the buffer
		 *
		 * @param node             the node to be formatted
		 * @param formatString     the format string to use
		 * @param isHeader         true indicates the formatString is a header format (vs regular content)
		 */
		void formatNode( const DSRDocumentTreeNode & node, const OFString & formatString, bool isHeader );

		/**
		 * Formats the specified node.
		 * Looks up configured rules for a matching rule and formats the content of the node accordingly.
		 * The formatted content will be appended to the buffer.
		 *
		 * @param node             the node to be formatted
		 */
		void formatNodeValue( const DSRDocumentTreeNode & node );


		/**
		 * Formats a String value and appends the formatted value to the buffer,
		 *
		 * @param value    the String value to be formatted
		 */
		void formatStringValue( const DSRStringValue & value );

		/**
		 * Formats a Coded Entry value and appends the formatted value to the buffer,
		 *
		 * @param value    the Coded Entry value to be formatted
		 */
		void formatCodedEntryValue( const DSRCodedEntryValue & value );

		/**
		 * Formats a Numeric Measurement value and appends the formatted value to the buffer,
		 *
		 * @param value    the Numeric Measurement value to be formatted
		 */
		void formatNumericMeasurementValue( const DSRNumericMeasurementValue & value );

		/**
		 * Formats a Spatial Coordinates value and appends the formatted value to the buffer,
		 *
		 * @param value    the Spatial Coordinates value to be formatted
		 */
		void formatSpatialCoordinatesValue( const DSRSpatialCoordinatesValue & value );

		/**
		 * Formats a Temporal Coordinates value and appends the formatted value to the buffer,
		 *
		 * @param value    the Temporal Coordinates value to be formatted
		 */
		void formatTemporalCoordinatesValue( const DSRTemporalCoordinatesValue & value );

		/**
		 * Formats a Composite Reference value and appends the formatted value to the buffer,
		 *
		 * @param value    the Composite Reference value to be formatted
		 */
		void formatCompositeReferenceValue( const DSRCompositeReferenceValue & value );

		/**
		 * Formats a Image Reference value and appends the formatted value to the buffer,
		 *
		 * @param value    the Image Reference value to be formatted
		 */
		void formatImageReferenceValue( const DSRImageReferenceValue & value );

		/**
		 * Formats a Waveform Reference value and appends the formatted value to the buffer,
		 *
		 * @param value    the Waveform Reference value to be formatted
		 */
		void formatWaveformReferenceValue( const DSRWaveformReferenceValue & value );

		/**
		 * Formats a Container value and appends the formatted value to the buffer,
		 *
		 * @param value    the Container value to be formatted
		 */
		void formatContainerContent( const DSRContainerTreeNode & value );


		/**
		 * Read header formats from the specified configuration file
		 * @param configFile  path to the configuration file
		 */
		void readHeaders( const OFString & configFile );

		/**
		 * Read header formats from the specified configuration files
		 * The first occurrence of a header specification shall be used
		 *
		 * @param configFiles  list of paths to the configuration files
		 */
		void readHeaders( const OFList< OFString > & configFiles );

		/**
		 * Read formatting rules from the specified configuration file
		 * @param configFile  path to the configuration file
		 */
		void readFormats( const OFString & configFile );

		/**
		 * Read formatting rules from the specified configuration files
		 * The first occurrence of a rule specification shall be used
		 *
		 * @param configFiles  list of paths to the configuration files
		 */
		void readFormats( const OFList< OFString > & configFiles );

		/**
		 * Parses concept name string read from a configuration file and sets the value of argument concept name accordingly
		 *
		 * @param concept        concept name string to parse
		 * @param conceptName    output parameter for the parsed values
		 */
		void parseConceptName( const OFString & concept, DSRCodedEntryValue & conceptName );

		/**
		 * Processes a format rule read from a configuration file.
		 * Creates a FormatRule from the strings received in the configuration file and adds the format rule to formatRules list.
		 *
		 * @param relType   relationship type string read from the configuration file
		 * @param valType   value type string read from the configuration file
		 * @param concept   concept name string read from the configuration file
		 * @param fmt       format string read from the configuration file
		 */
		void processFormat( OFString & relType, OFString & valType, OFString & concept, OFString & fmt );

		/**
		 * Reads the next token from the buffer.
		 * Finds the next (not quoted) separator in the buffer, and sets the value of the token argument.
		 * buffer will point to the next character to process.
		 *
		 * @param buffer      the buffer to get the token from, will be modified to point to the next unprocessed character
		 * @param separator   token separator character
		 * @param token       output parameter for the token parsed
		 *
		 * @return true if and only if separator character was successfully found
		 */
		bool readNextToken( const char *& buffer, char separator, OFString & token );

		/**
		 * Skips whitespaces in the character buffer.
		 *
		 * @param buffer  in/out parameter. Will point to the first not whitespace character
		 */
		void skipSpaces( const char *& buffer );

		/**
		 * Removes trailing white spaces from a string.
		 *
		 * @param s   in/out parameter for the string to trim
		 * @return the same as s
		 */
		OFString & rtrim( OFString & s );


		/**
		 * Sets header format string. If header format for the particular level is already set, it will not be modified.
		 *
		 * @param level   header level to set
		 * @param format  format string for the header
		 */
		void addHeaderFormat( size_t level, const OFString & format );

		/**
		 * Sets default header format string. If default header format is already set, it will not be modified.
		 *
		 * @param format  format string for the default header
		 */
		void setDefaultHeaderFormat( const OFString & format );

		/**
		 * Increase level (depth in the content tree).
		 */
		inline void increaseLevel() {
			m_level++;
		}

		/**
		 * Decrease level (depth in the content tree).
		 */
		inline void decreaseLevel() {
			m_level--;
		}


		/**
		 * Add a rule to the format rule list.
		 * If a matching rule is already specified, it will not be modified.
		 *
		 * @param rule the rule to add
		 */
		void addRule( FormatRule * rule );


		/**
		 * Return header format according to the current level in the content tree
		 *
		 * @return reference to the header format
		 */
		const OFString & getHeaderFormat() const;



	public:

		/**
		 * Default constructor to create a new Formatter object.
		 * Formatted content will be added to buffer
		 *
		 * @param buffer   write formatted content to this buffer
		 */
		Formatter( Buffer & buffer );

		/**
		 * Destructor to destroy this Formatter object.
		 */
		virtual ~Formatter();

		/**
		 * Read formatting rules for the specified SR concept name and target format
		 *
		 * @param configDirectories  list of configuration directories to search for formatting rules
		 * @param rootConceptName    Concept Name of the SR content tree
		 * @param targetName         Formatting target (External, Findings, ...)
		 */
		void readRules( const OFList< OFString > & configDirectories, const DSRCodedEntryValue & rootConceptName, const OFString & targetName );

		/**
		 * Formats the specified node according to the formatting rules.
		 * The formatted content will be appended to the buffer.
		 *
		 * @param node             the node to be formatted
		 */
		void appendFormattedNode( const DSRDocumentTreeNode & node );


		/**
		 * Sets the content tree of the SR object
		 *
		 * @param tree   content tree of the SR document
		 */
		inline void setTree( DSRDocumentTree * tree ) {
			m_tree = tree;
		}

		/**
		 * Sets verbose mode.
		 * In verbose mode, messages about reading rules from configuration files will be printed to the standard output.
		 *
		 * @param value   true indicates to print verbose messages
		 */
		inline void setVerbose( bool value ) {
			m_verbose = value;
		}


		/**
		 * Prints formatting rules to the specified output stream
		 *
		 * @param out   output stream to write to
		 */
		const void dumpRules( std::ostream & out );
};

}

#endif /* _IM_SR_FORMATTER_H_ */
