/*
 * Component: iface
 *
 * Description:
 *   Header file for order matching configuration classes
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __ORDER_MATCHING_CONFIG_H_INCLUDED__
#define __ORDER_MATCHING_CONFIG_H_INCLUDED__

#include <dcmtk/ofstd/ofstring.h>
#include <dcmtk/ofstd/oflist.h>
#include <dcmtk/dcmdata/dctk.h>

#include "dbim.h"

/**
 * Holds information on Dicom Tags and database field names of data
 * indexed to the database
 */
class TagField {
	private:

		/** DICOM tag of field */
		DcmTagKey m_tag;

		/** SQL field name */
		OFString m_dbField;

		/** Indicates whether the field is type of CHAR */
		OFBool m_isChar;

	public:

		/**
		 * Constructor for creating a new TagField object.
		 *
		 * @param tag         DICOM tag of the field
		 * @param dbField     SQL field name
		 * @param isChar      Indicates whether the field is type of CHAR
		 */
		TagField( const DcmTagKey & tag, const OFString & dbField, OFBool isChar );

		/**
		 * Copy constructor
		 *
		 * @param orig     original TagField object to be copied
		 */
		TagField( const TagField & orig );

		/**
		 * Destructor for destroying TagField object
		 */
		~TagField();

		/**
		 * Retrieves the DICOM tag of the field
		 *
		 * @return the DICOM tag of the field
		 */
		inline const DcmTagKey getTag() const {
			return m_tag;
		}

		/**
		 * Retrieves the SQL field name
		 *
		 * @return the SQL field name
		 */
		inline const OFString getDbField() const {
			return m_dbField;
		}

		/**
		 * Retrieves whether the SQL field is type of CHAR or not
		 *
		 * @return OFtrue  if the SQL field is type of CHAR
		 *         OFFalse otherwise
		 */
		inline OFBool isChar() const {
			return m_isChar;
		}

		/**
		 * Assignment operator
		 *
		 * param orig    TagField object to be copied
		 *
		 * @return reference to this
		 */
		TagField & operator=( const TagField & orig );
};

/**
 * Holds information on an entry in the order matching config file.
 * An order matching entry contains list of fields to be used to identify an
 * incoming order from a device.
 */
class MatchingEntry {
		friend class MatchingConfig;
	private:

		/** Device ID */
		OFString m_device;

		/** List of fields to identify an order */
		OFList< TagField > m_fields;

		/** SQL connection handle */
		DB_Query & m_sqlHandle;

		/**
		 * SQL encodes a string value to be included in a query.
		 * Will be put escape character (backslash: '\') in front of quotes and backslashes
		 *
		 * @param str   String value to be SQL encoded
		 *
		 * @return the SQL encoded string
		 */
		OFString sqlEncode( const OFString & str );

	protected:

		/**
		 * Assignment operator.
		 *
		 * Will be used by MatchingConfig::checkAndLoadConfig() to override default settings
		 */
		MatchingEntry & operator=( const MatchingEntry & orig );

	public:

		/**
		 * Constructor for creating a new MatchingEntry object
		 *
		 * @param sqlHandle  SQL connection handle
		 * @param device     Device ID
		 */
		MatchingEntry( DB_Query & sqlHandle, const OFString & device );

		/**
		 * Destructor for destroying MatchingEntry object
		 */
		~MatchingEntry();

		/**
		 * Adds a field to the field list.
		 *
		 * @param tagField  field to be added to the list
		 */
		inline void addDbField( const TagField & tagField ) {
			m_fields.push_back( tagField );
		}

		/**
		 * Retrieves the Device ID
		 *
		 * @return the Device ID
		 */
		inline const OFString & getDevice() const {
			return m_device;
		}

		/**
		 * Checks if the field list is empty
		 *
		 * @return OFTrue if the field list is empty, OFFalse otherwise
		 */
		inline OFBool isEmptyFieldList() const {
			return m_fields.empty();
		}

		/**
		 * Gets field information from Dcfields and Dctypes sql table and add the field
		 * to the field list.
		 *
		 * @param field    SQL field name of the field to be added to the field list
		 *
		 * @return true on success, false otherwise
		 *
		 * @see #addDbField(const TagField &)
		 */
		bool addField( const OFString & field );

		/**
		 * Creates where clause from data in dataset to find matches.
		 * The dataset is expected to be the order dataset received from the device.
		 * Field values will be read from dataset for configured fields,
		 * and a SQL where clause for an SQL query will be created using the field values.
		 *
		 * @param dataset      order dataset received from the device
		 * @param whereClause  output parameter for the generated where clause
		 *
		 * @throw IMException if an error occurs
		 */
		void createWhereClause( DcmDataset * dataset, OFString & whereClause );

		/**
		 * Gets Study Instance UID for an order received from device.
		 * The dataset is expected to be the order dataset received from the device.
		 * Field values will be read from dataset for configured fields,
		 * and a SQL query will be created using the field values. If there is exactly
		 * one match in the database for the field values, the study uid of the matching
		 * database entry will be retrieved.
		 *
		 * @param dataset      order dataset received from the device
		 * @param studyIUID    output parameter for the Study Instance UID of the order
		 *
		 * @return true if and only if there was exactly one match and Study Instance UID could be specified
		 */
//	bool getStudyIUID(DcmDataset * dataset, char *& studyIUID);
};

/**
 * Holds configuration of order matching.
 * Contains list of fields to be used getting Study Instance UID from database for each device.
 * In case of devices not listed in this config the default matching (Accession Number) will be used.
 */
class MatchingConfig {
	private:

		/** name of the configuration file to be read */
		OFString m_configFileName;

		/** Last modification time of the config file */
		time_t m_lastModified;

		/** Default matching entry */
		MatchingEntry m_defaultEntry;

		/**
		 * Skips whitespaces in the string. Will not modify the original string
		 * but will refer it.
		 *
		 * @param str   skip whitespaces in this string
		 *
		 * @return the pointer of the first non whitespace character in str
		 */
		char * skipWhiteSpaces( char * str );

		/**
		 * Removes leading and trailing whitespaces from the string. The original string might be modified,
		 * the first whitespace character after the last non whitespase character will be replaced by EOS.
		 *
		 * @param   string to be trimmed
		 *
		 * @return the pointer of the first non whitespace character in str
		 */
		char * trim( char * str );

		/**
		 * Replaces all occurrences of a character in the string. The original string will be modified.
		 *
		 * @param str   string to be modified
		 * @param src   character to be replaced
		 * @param dest  replacement character
		 */
		void replace( char * str, char src, char dest );

		/**
		 * Retrieves the next value from the string delimited by delim.
		 * The retrieved value will be trimmed.
		 *
		 * @param str     search for value in this string. If value was detected, the value of this parameter
		 *                will be modified to the first character after delim.
		 * @param delim   value delimiter
		 *
		 * @return the next value detected, or NULL if delimiter was not found
		 */
		char * getNextValue( char *& str, char delim );

		/**
		 * Clears all configured entries.
		 */
		void clearEntries();

	protected:

		/** SQL connection handle */
		DB_Query m_sqlHandle;

		/** List of configured entries */
		OFList< MatchingEntry * > m_entries;

		/**
		 * Checks if the config file was modified since it was loaded and reload it.
		 *
		 * @return true if the config file was successfully (re)loaded, false otherwise
		 */
		virtual bool checkAndLoadConfig();

		/**
		 * Find config entry for a device. If there is no config entry for the device,
		 * then the default entry will be returned.
		 *
		 * @param device       Device ID of the device
		 *
		 * @return config entry to be used for matching for orders received from this device
		 */
		MatchingEntry & findEntryForDevice( const OFString & device );

	public:

		/**
		 * Default constructor for creating MatchingConfig entry
		 * The created MatchingConfig object will always use default matching
		 */
		MatchingConfig();

		/**
		 * Constructor for creating MatchingConfig entry
		 * Matching rules will be read from the config file.
		 * The configuration file should look like:
		 * <device>="<field>[,<field>]"
		 * empty and comment lines (starting with '#') will be ignored.
		 *
		 * @param configFileName   full path to the configuration file
		 */
		MatchingConfig( const OFString & configFileName );

		/**
		 * Sets matching configuration file .
		 * Matching rules will be read from the config file.
		 * The configuration file should look like:
		 * <device>="<field>[,<field>]"
		 * empty and comment lines (starting with '#') will be ignored.
		 *
		 * @param configFileName   full path to the configuration file
		 */
		inline void setConfigFile( const OFString & configFileName ) {
			m_configFileName = configFileName;
		}

		/**
		 * Destructor for destroying MatchingConfig object
		 */
		virtual ~MatchingConfig();
};

/**
 * Structure returning information on a matching study
 */
class MatchingStudyEntry {
		friend class OrderMatchingConfig;
		friend class ReportMatchingConfig;
		/** UID of the study */
		OFString m_study;
		/** Proposed study UID for orders */
		OFString m_proposedStudy;
		/** Deleted flag */
		OFBool m_isDeleted;
		/** Order flag */
		OFBool m_isOrder;
		/** True indicates there are multiple matches for study identifiers */
		OFBool m_hasMultipleMatch;

	protected:

		/**
		 * Sets study UID
		 *
		 * @param study new value for study UID
		 */
		inline void setStudy( const char * study ) {
			m_study = study;
		}

		/**
		 * Sets proposed study UID
		 *
		 * @param proposedStudy   new value for proposed study UID
		 */
		inline void setProposedStudy( const char * proposedStudy ) {
			m_proposedStudy = proposedStudy;
		}

		/**
		 * Sets Deleted flag
		 *
		 * @param isDeleted   new value for the deleted flag
		 */
		inline void setDeletedFlag( const OFBool & isDeleted ) {
			m_isDeleted = isDeleted;
		}

		/**
		 * Sets Order flag
		 *
		 * @param isOrder   new value for the order flag
		 */
		inline void setOrderFlag( const OFBool & isOrder ) {
			m_isOrder = isOrder;
		}

		/** clear current values and sets has multiple matches flag*/
		void setMultipleMatch();

	public:
		/**
		 * Constructor creating a matching study entry object
		 */
		MatchingStudyEntry();

		/**
		 * @return study instance UID
		 */
		inline const OFString & getStudy() const {
			return m_study;
		}

		/**
		 * @return proposed study instance UID
		 */
		inline const OFString & getProposedStudy() const {
			return m_proposedStudy;
		}

		/**
		 * @return OFTrue if the study entry is in deleted state
		 */
		inline OFBool isDeleted() const {
			return m_isDeleted;
		}

		/**
		 * @return OFTrue if the study entry is an order
		 */
		inline OFBool isOrder() const {
			return m_isOrder;
		}

		/**
		 * @return OFTrue if there are multiple matching studies/orders
		 */
		inline OFBool hasMultipleMatch() const {
			return m_hasMultipleMatch;
		}
};

class OrderMatchingConfig : public MatchingConfig {
	public:

		/**
		 * Default constructor for creating OrderMatchingConfig entry
		 * The created OrderMatchingConfig object will always use default matching
		 */
		OrderMatchingConfig();

		/**
		 * Constructor for creating OrderMatchingConfig entry
		 * Matching rules will be read from the config file.
		 * The configuration file should look like:
		 * <device>="<field>[,<field>]"
		 * empty and comment lines (starting with '#') will be ignored.
		 *
		 * @param configFileName   full path to the configuration file
		 */
		OrderMatchingConfig( const OFString & configFileName );

		/**
		 * Destructor for destroying MatchingConfig object
		 */
		virtual ~OrderMatchingConfig();

		/**
		 * Gets Study Instance UID for an order received from the device.
		 * The following steps will be performed:
		 * 1./ If the dataset contains Study Instance UID, then it will be returned.
		 * 2./ If there is a configured entry for the device, the Study Instance UID
		 *     will be get using the configured rules.
		 * 3./ If there was no configured entry for the device, the default matching
		 *     will be performed (Matching by Accession Number)
		 *
		 * @param device       Device ID of the device the order was received from
		 * @param dataset      Order dataset received
		 * @param studyEntry   Output parameter for the study and proposed study UIDs
		 *
		 * @return true if and only if Study Instance UID was successfully retrieved
		 *
		 * @throws IMException if an error occurred (SQL error, no identifier field in the dataset, ...)
		 */
		bool getStudyIUID( const OFString & device, DcmDataset * dataset, MatchingStudyEntry & studyEntry );

	private:

		/**
		 * Checks if the config file was modified since it was loaded and reload it.
		 * Overrides checkAndLoadConfig() in MatchinConfig, adds built in default entry for the scheduler module (SCHEDULER).
		 *
		 * @return true if the config file was successfully (re)loaded, false otherwise
		 */
		virtual bool checkAndLoadConfig();
};

class OrderGroupingConfig : public MatchingConfig {
	public:

		/**
		 * Default constructor for creating OrderGroupingConfig entry
		 * The created OrderGroupingConfig object will always use default merging
		 */
		OrderGroupingConfig();

		/**
		 * Constructor for creating OrderGroupingConfig entry
		 * Grouping rules will be read from the config file.
		 * The configuration file should look like:
		 * <device>="<field>[,<field>]"
		 * empty and comment lines (starting with '#') will be ignored.
		 *
		 * @param configFileName   full path to the configuration file
		 */
		OrderGroupingConfig( const OFString & configFileName );

		/**
		 * Destructor for destroying OrderGroupingConfig object
		 */
		virtual ~OrderGroupingConfig();

		/**
		 * Checks if there's a Proposed Study Instance UID for an incoming order and generate one if necessary.
		 * Generated Proposed Study Instance UID will be put into the order dataset.
		 * Proposed Study Instance UID is used to merge incoming orders into one study
		 *
		 * @param device       Device ID of the device the order was received from
		 * @param dataset      Order dataset received (Generated Proposed
		 * @param studyIUID    Study Instance UID of the incoming order
		 *
		 * @return true if Proposed Study Instance UID was found or has been successfully created
		 */
		bool createAndAddProposedStudyIUID( const OFString & device, DcmDataset * dataset, const char * studyIUID );
};

class ReportMatchingConfig : public MatchingConfig {

	public:

		/**
		 * Default constructor for creating a new ReportMatchingConfig object.
		 * The created ReportMatchingConfig object will use default matching
		 */
		ReportMatchingConfig();

		/**
		 * Constructor for creating a new ReportMatchingConfig object.
		 * Matching rules will be read from the config file.
		 * The configuration file should look like:
		 * <device>="<field>[,<field>]"
		 * empty and comment lines (starting with '#') will be ignored.
		 *
		 * @param configFileName   full path to the configuration file
		 */
		ReportMatchingConfig( const OFString & configFileName );

		/**
		 * Destructor to destroy this ReportMatchingConfig object
		 */
		virtual ~ReportMatchingConfig();

		/**
		 * Gets Study Instance UID for a report received from the device.
		 *  //The following steps will be performed:
		 *  // 1./ If the dataset contains Study Instance UID, then it will be returned.
		 *  // 2./ If there is a configured entry for the device, the Study Instance UID
		 *  //     will be get using the configured rules.
		 *  // 3./ If there was no configured entry for the device, the default matching
		 *  //     will be performed (Matching by Accession Number)
		 *
		 * @param device       Device ID of the device the order was received from
		 * @param dataset      Order dataset received
		 * @param studyEntry   Output parameter for the study and proposed study UIDs
		 *
		 * @return true if and only if Study Instance UID was successfully retrieved
		 *
		 * @throws IMException if an error occurred (SQL error, no identifier field in the dataset, ...)
		 */
		bool getStudyIUID( const OFString & device, DcmDataset * dataset, MatchingStudyEntry & studyEntry );

};

#endif // __ORDER_MATCHING_CONFIG_H_INCLUDED__
